from vlab4mic import experiments, sweep_generator
import numpy as np
import pytest
import copy


def test_run_parameter_sweep():
    sweep_gen_test = sweep_generator.run_parameter_sweep(
        structures=["7R5K",],
        probe_templates=["NPC_Nup96_Cterminal_direct",],
        sweep_repetitions=3,
        # parameters for sweep
        labelling_efficiency=(0,1,1),
        structural_integrity=[0,0.5, 1],
        structural_integrity_small_cluster=[300,],
        structural_integrity_large_cluster=[600,],
        #exp_time=[0.001, 0.01,],
        # output and analysis
        output_name="vlab_script",
        return_generator=True,
        save_sweep_images=False,
        save_analysis_results=False,
        run_analysis=True,
        analysis_plots=False
        )
    
    assert sweep_gen_test.analysis["dataframes"] is not None
    param_groups = list(sweep_gen_test.params_by_group.keys())
    total_combinations = 1
    for group_name in param_groups:
        if len(sweep_gen_test.params_by_group[group_name]) > 0:
            for param_name in sweep_gen_test.params_by_group[group_name]:
                total_combinations *= len(
                    sweep_gen_test.params_by_group[group_name][param_name]
                )
    vsamples_unique_ids = len(sweep_gen_test.virtual_samples_parameters.keys())
    assert total_combinations == vsamples_unique_ids

def test_custom_metric():
    
    def mean_value(reference_image = None,
            reference_image_pixelsize_nm = None,
            simulated_image = None,
            simulated_image_pixelsize_nm = None,
            image_mask = None,
            resized_reference_image = None,
            resized_simulated_image = None,
            *args,**kwargs):
        return np.mean(simulated_image)

    
    sweep_gen = sweep_generator.run_parameter_sweep(
        sweep_repetitions=3,
        # parameters for sweep
        labelling_efficiency=(0, 1, 0.5),  # values between 0 and 1 with step of 0.5
        return_generator=True,
        analysis_plots=False,
        save_sweep_images=False,  # By default, the saving directory is set to the home path of the user
        save_analysis_results=False,
        run_analysis=True,
        custom_metrics=[mean_value,],
    )


import numpy as np
from .points_transforms import rotate_point


def normals_by_scaling(epitope_locs, scale=0.95):
    scaled_epitopes_locs = coordinates_scaling(epitope_locs, scale)
    # calculate the normals of the epitopes
    # note that this approach will work for convex shapes
    # this assumes as well the scaled version is smaller
    normals = np.zeros((len(epitope_locs), 3))
    for i in range(len(epitope_locs)):
        normals[i, :] = epitope_locs[i] - scaled_epitopes_locs[i]
    # then the points from which the normals are traced are the epitopes themselves
    # normals_ft_epitopes = [normals, epitope_locs]
    return normals
    # return normals_ft_epitopes

def global_normal_direction(epitope_locs, normal_vector = np.array([0,0,1])):
    normals = np.zeros((len(epitope_locs), 3))
    for i in range(len(epitope_locs)):
        normals[i, :] = normal_vector
    return normals

def coordinates_scaling(
    epitope_array, scaling_factor
):  ## formerly named epitopes_scaling
    # epitope_array is a numpy array epitope locations
    # first get the centroid of the epitope array
    # (this input can be the averale location of the atoms in the epitopes)
    centroid0 = np.mean(epitope_array, axis=0)
    # scale down epitopes
    naive_scaled = epitope_array * scaling_factor  # for now this number can be fixed
    # get the centroid of the scaled epitope array
    centroid1 = np.mean(naive_scaled, axis=0)  # this is not necesary for now
    naive_scaled_centered = np.zeros(naive_scaled.shape)
    displacement = centroid1 - centroid0
    for i in range(naive_scaled.shape[0]):
        naive_scaled_centered[i] = naive_scaled[i] - displacement
    return naive_scaled_centered  # output is centered at the centroid of the input


def add_wobble(pivot, direction, cone_angle_deg=30, length=1):
    """
    Adds a random wobble to the given vector within a cone defined by cone_angle_deg,
    with uniform distribution over the cone's surface area.
    The wobble is relative to the pivot and the direction of the vector.

    :param pivot: (np.array) The pivot point in 3D space
    :param direction: (np.array) The unit vector representing the initial direction from the pivot.
    :param cone_angle_deg: (float) The maximum angle of the cone (in degrees).
    :param length: (float) The length of the translation vector to preserve after wobbling.

    :return: (np.array) The new wobbling direction vector.
    """
    # Ensure direction is a unit vector
    direction = direction / np.linalg.norm(direction)

    # Convert the cone angle from degrees to radians
    cone_angle_rad = np.radians(cone_angle_deg)

    # Sample azimuthal angle niformly
    phi = np.random.uniform(0, 2 * np.pi)

    # The cone defines a boundary for theta angles to choose from
    cos_theta_max = np.cos(cone_angle_rad)
    # this max theta is the angle between the originall vector
    # and the new one, which is to be place within a cone
    # then this max theta is actually a lower bound
    cos_theta = np.random.uniform(cos_theta_max, 1)
    # but this is the arc in radians, so we get its arcosine
    theta = np.arccos(cos_theta)

    # Random wobble vector in spherical coordinates
    # note that this is the relative wobble to a unitary vector
    x_wobble = np.sin(theta) * np.cos(phi)
    y_wobble = np.sin(theta) * np.sin(phi)
    z_wobble = np.cos(theta)

    # Build the unitary wobble vector
    wobble_vector = np.array([x_wobble, y_wobble, z_wobble])

    # Now, align the wobble vector to the cone defined by the initial direction
    # Find the axis of rotation (cross product of the initial direction and the z-axis)
    rotation_axis = np.cross(np.array([0, 0, 1]), direction)

    # If the direction is already aligned with the z-axis, pick a rotation axis perpendicular to it
    if np.linalg.norm(rotation_axis) < 1e-6:
        rotation_axis = np.cross(np.array([1, 0, 0]), direction)

    # Rotate the wobble vector around the axis defined by the initial direction
    angle = np.arccos(
        np.dot(np.array([0, 0, 1]), direction)
    )  # Angle between the z-axis and the direction vector
    rotated_wobble = rotate_point(wobble_vector, rotation_axis, angle)

    # Adjust the resulting vector to the desired length (this is the wobble vector)
    wobble_translation = rotated_wobble * length
    # Add the pivot and the translation to get the endpoint
    wobble_endpoint = pivot + wobble_translation
    # Return the wobble as an endpoint relative to the pivot
    return wobble_endpoint



import numpy as np


def elliptical_gaussian_3sigmas(shape, std_devs):
    """
    Create a 3D elliptical Gaussian array.

    Parameters:
    - shape: Tuple, shape of the array (depth, height, width).
    - std_dev: Tuple, standard deviations along each axis (z, y, x).

    Returns:
    - ndarray: 3D NumPy array representing the elliptical Gaussian.
    """
    center = tuple(dim // 2 for dim in shape)
    z, y, x = np.ogrid[: shape[0], : shape[1], : shape[2]]
    elliptical_gaussian = np.exp(
        -(
            (z - center[0]) ** 2 / (2 * std_devs[0] ** 2)
            + (y - center[1]) ** 2 / (2 * std_devs[1] ** 2)
            + (x - center[2]) ** 2 / (2 * std_devs[2] ** 2)
        )
    )
    elliptical_gaussian = elliptical_gaussian / np.sum(elliptical_gaussian)
    return elliptical_gaussian



from vlab4mic import workflows
from vlab4mic.experiments import image_vsample
import pytest
import copy

structure_list = [
    "2RCJ",
]


@pytest.mark.parametrize("structure_id", structure_list)
def test_load_structure(structure_id, configuration_directory):
    configuration_path = configuration_directory
    structure, structure_param = workflows.load_structure(
        structure_id, configuration_path
    )
    assert structure.assembly_refpt.shape == (3,)


# test user-input file


def test_structure_normals(experiment_7r5k_base):
    structure = copy.deepcopy(experiment_7r5k_base.structure)
    structure.assign_normals2targets()
    #structure.show_target_labels(with_normals=True, show_axis=True)


def test_symmetry_parsing():
    images, noiseless, experiment = image_vsample(
        run_simulation=False,
        clear_experiment=True,
    )
    assert experiment.structure.assymetric_defined == True


import numpy as np


def format_coordinates(
    coords, plotsize=1, plotalpha=1, plotmarker="o", plotcolour="k", name="", **kwargs
):
    formatted = {
        "coordinates": coords,
        "plotsize": plotsize,
        "plotalpha": plotalpha,
        "plotmarker": plotmarker,
        "plotcolour": plotcolour,
        "label_name": name,
    }
    for key, value in kwargs.items():
        formatted[key] = value
    return formatted


def set_colorplot(plotting_params):
    # this function takes the dictionary that is stored in self.label_targets
    if len(plotting_params) == 0:
        defaultcolour = "#984ea3"
    else:
        unavailable = []
        for labelname in plotting_params.keys():
            # print(f'colours taken: {plotting_params[labelname]["plotcolour"]}')
            unavailable.append(plotting_params[labelname]["plotcolour"])
        CB_color_cycle = [
            "#377eb8",
            "#ff7f00",
            "#4daf4a",  # https://gist.github.com/thriveth/8560036
            "#f781bf",
            "#a65628",
            "#984ea3",
            "#999999",
            "#e41a1c",
            "#dede00",
        ]
        defaultcolour = "#984ea3"
        while defaultcolour in unavailable:
            defaultcolour = np.random.choice(CB_color_cycle)
    return defaultcolour



from ..experiments import ExperimentParametrisation
from ..utils.io.yaml_functions import load_yaml
import os
import vlab4mic
from . import metrics
import numpy as np
import pandas as pd
import itertools
import copy
from tqdm import tqdm
from IPython.utils import io


def sweep_vasmples(
    experiment: ExperimentParametrisation = None,
    structures=None,
    probes=None,
    probe_parameters=None,
    particle_structural_integrity=None,
    virtual_samples=None,
    repetitions=1,
    use_experiment_structure=False,
    **kwargs,
):
    """
    Generate virtual samples for all combinations of structures, probes, probe parameters, structural_integrity, and virtual sample parameters.

    Parameters
    ----------
    experiment : ExperimentParametrisation, optional
        Experiment object to use. If None, a new one is created.
    structures : list, optional
        List of structure IDs.
    probes : list, optional
        List of probe names.
    probe_parameters : dict, optional
        Dictionary of probe parameter sets.
    particle_structural_integrity : dict, optional
        Dictionary of structural integrity parameter sets.
    virtual_samples : dict, optional
        Dictionary of virtual sample parameter sets.
    repetitions : int, optional
        Number of repetitions for each combination.
    **kwargs
        Additional keyword arguments.

    Returns
    -------
    experiment : ExperimentParametrisation
        The experiment object used.
    vsample_outputs : dict
        Dictionary of generated virtual samples.
    vsample_params : dict
        Dictionary of parameter combinations for each sample.
    """
    # empty lists, but fill up with default
    pck_dir = os.path.dirname(os.path.abspath(vlab4mic.__file__))
    local_dir = os.path.join(pck_dir, "configs")
    default_probe = "NHS_ester"
    default_fluorophore = "AF647"
    if experiment is None:
        experiment = ExperimentParametrisation()
        experiment.clear_experiment()
    if structures is None:
        structures = [
            "1XI5",
        ]
    if probes is None:
        print("Probe list is None. Using default probe")
        probes = [
            default_probe,
        ]
    if probe_parameters is None:
        probe_filepath = os.path.join(
            local_dir, "probes", default_probe + ".yaml"
        )
        default_params = load_yaml(probe_filepath)
        probe_parameters = dict()
        probe_parameters[0] = None
    if particle_structural_integrity is None:
        particle_structural_integrity = dict()
        particle_structural_integrity[0] = {"use_structural_integrity": False}
    if virtual_samples is None:
        # vprobe_filepath = os.path.join(local_dir, "probes", default_probe + ".yaml")
        # default_vsample =  load_yaml(vprobe_filepath)
        virtual_samples = dict()
        virtual_samples[0] = {"random_orientations": False}
        # virtual_samples = [
        #    None,
        # ]
    vsample_params = dict()
    vsample_outputs = dict()
    if "relative_positions" in experiment.virtualsample_params.keys():
        relative_positions = copy.deepcopy(
                experiment.virtualsample_params["relative_positions"]
                )
        relative_positions_exist = True
    else:
        relative_positions_exist = False
        relative_positions = None

    for struct in structures:
        if use_experiment_structure:
            experiment.build(modules=["structure",])
        else:
            experiment.structure_id = struct
            experiment._build_structure()
        # for rep in range(repetitions):
        probe_n = 0
        for probe in probes:
            print(f"probe: {probe}")
            # probe_param_n = 0
            for probe_param_n, p_param in probe_parameters.items():
                # copy p_param?
                experiment.remove_probes()
                if p_param is None:
                    experiment.add_probe(
                        probe_template=probe,
                    )
                    # experiment.probe_parameters[probe] = default_params
                else:
                    p_param_copy = copy.deepcopy(p_param)
                    p_param_copy["fluorophore_id"] = default_fluorophore
                    experiment.add_probe(probe_template=probe, **p_param_copy)
                for structural_integrity_n, structural_integrity_pars in particle_structural_integrity.items():
                    particle_structural_integrity_copy = copy.deepcopy(structural_integrity_pars)
                    for d_key, d_val in particle_structural_integrity_copy.items():
                        if d_key == "structural_integrity_small_cluster":
                            experiment.structural_integrity_eps["eps1"] = d_val
                        if d_key == "structural_integrity_large_cluster":
                            experiment.structural_integrity_eps["eps2"] = d_val
                        if d_key == "structural_integrity":
                            experiment.structural_integrity_eps["structural_integrity"] = d_val
                    print(experiment.probe_parameters)
                    experiment._build_particle(keep=True)
                    if experiment.generators_status("particle"):
                        if len(experiment.particle.emitters) == 0:
                            print(
                                f"Skipping {probe}. No emitters were generated"
                            )
                            break
                    else:
                        break
                    # vsample_n = 0
                    for vsample_n, vsample_pars in virtual_samples.items():
                        _exported_field = None
                        # combination += str(vsample_n)
                        #experiment.set_virtualsample_params(update_mode=True, **vsample_pars)
                        # _exported_field = experiment._build_coordinate_field(
                        #    keep=False, use_self_particle=True, **vsample_pars
                        # )
                        for rep in range(repetitions):
                            if relative_positions_exist and relative_positions is not None: # keep the same positions
                                experiment.clear_virtual_sample()
                                experiment.set_virtualsample_params(
                                    update_mode=True,
                                    particle_positions = relative_positions,
                                    **vsample_pars)
                                experiment.build(
                                    modules=["coordinate_field"],
                                    use_self_particle=True,
                                )
                            else:
                                experiment.clear_virtual_sample()
                                experiment.set_virtualsample_params(
                                    update_mode=True,
                                    **vsample_pars)
                                experiment.build(
                                    modules=["coordinate_field"],
                                    use_self_particle=True,
                                )
                                relative_positions = copy.deepcopy(
                                    experiment.coordinate_field.molecules_params["relative_positions"]
                                    )
                                relative_positions_exist = True
                            _exported_field = (
                                experiment.coordinate_field.export_field()
                            )
                            combination_n = (
                                str(probe_n)
                                + "_"
                                + str(probe_param_n)
                                + "_"
                                + str(structural_integrity_n)
                                + "_"
                                + str(vsample_n)
                            )
                            _parameters = [
                                struct,
                                probe,
                                p_param,
                                structural_integrity_pars,
                                vsample_pars,
                            ]
                            if combination_n not in vsample_params.keys():
                                vsample_params[combination_n] = _parameters
                                vsample_outputs[combination_n] = []
                            vsample_outputs[combination_n].append(
                                _exported_field
                            )
            probe_n += 1
    return experiment, vsample_outputs, vsample_params


def sweep_modalities_updatemod(
    experiment: ExperimentParametrisation = None,
    vsample_outputs=None,
    vsampl_pars=None,
    modalities: list = None,
    modality_params: dict = None,
    modality_acq_prams: dict = None,
):
    """
    Simulate image acquisitions for all virtual samples and modalities.

    Parameters
    ----------
    experiment : ExperimentParametrisation, optional
        Experiment object to use. If None, a new one is created.
    vsample_outputs : dict, optional
        Dictionary of virtual sample outputs.
    vsampl_pars : dict, optional
        Dictionary of virtual sample parameters.
    modalities : list, optional
        List of modality names.
    modality_params : dict, optional
        Dictionary of modality parameter sets.
    modality_acq_prams : dict, optional
        Dictionary of modality acquisition parameter sets.

    Returns
    -------
    experiment : ExperimentParametrisation
        The experiment object used.
    mod_outputs : dict
        Dictionary of simulated image outputs.
    mod_params : dict
        Dictionary of parameter combinations for each modality.
    pixelsizes : dict
        Dictionary of pixel sizes for each modality.
    """
    default_mod = "Confocal"
    default_aqc = dict(nframes=2, exp_time=0.005)
    default_vsample = "None"
    mod_outputs = dict()
    mod_outputs_masks = dict()
    mod_params = dict()
    if experiment is None:
        experiment = ExperimentParametrisation()
    if vsample_outputs is None:
        vsample_outputs = [
            default_vsample,
        ]
        # needs default sample. can be a minimal field with a single emitter
    experiment.clear_modalities()
    if modalities == "all":
        list_of_locals = ["Widefield", "Confocal", "SMLM", "STED"]
        for local_mode in list_of_locals:
            experiment.add_modality(local_mode)
    elif modalities is None:
        experiment.add_modality("STED")
    else:
        for modality_name in modalities:
            experiment.add_modality(modality_name)
    if modality_params is None:
        modality_params = {}
        modality_params[0] = {}
    if modality_acq_prams is None:
        modality_acq_prams = {}
        modality_acq_prams[0] = None
    experiment._build_imager(use_local_field=False, prints=False)
    with io.capture_output() as captured:
        experiment._gen_modality_noise_images()
    pixelsizes = dict()
    imager_scale = experiment.imager.roi_params["scale"]
    scalefactor = np.ceil(imager_scale / 1e-9)  # in nanometers
    for mod_name, parameters in experiment.imaging_modalities.items():
        pixelsizes[mod_name] = np.array(
            parameters["detector"]["pixelsize"] * scalefactor
        )
    for vsmpl_id in tqdm(
        list(vsampl_pars.keys()),
        position=0,
        leave=True,
        desc="Unique parameter combination",
    ):
        for virtualsample in tqdm(
            list(vsample_outputs[vsmpl_id]),
            position=1,
            leave=False,
            desc="Repeats",
        ):
            experiment.imager.import_field(**virtualsample)
            with io.capture_output() as captured:
                mod_n = 0
                for modality_name in experiment.selected_mods.keys():
                    for mod_pars_number, mod_pars in modality_params.items():
                        print(f"Modality: {modality_name}, params set: {mod_pars} ")
                        experiment.update_modality(modality_name, **mod_pars)
                        # calculate virtual sample mask per modality
                        vsample_mask_per_mod = experiment.imager.generate_modality_mask(modality=modality_name)
                        #mod_threshold = 1  # needs to be modality specific
                        for (
                            mod_acq_number,
                            acq_pars,
                        ) in modality_acq_prams.items():
                            if acq_pars:
                                # expects to have 'channels' key with one channel at least
                                if "channels" not in acq_pars.keys():
                                    channel = "ch0"
                                    acq_pars = {"channels": [channel,]}
                                experiment.set_modality_acq(
                                        modality_name=modality_name, **acq_pars
                                    )
                            else:
                                channel = "ch0"
                                experiment.set_modality_acq(
                                    modality_name=modality_name, channels = [channel,]
                                )
                                acq_pars = {"channels": [channel,]}
                            # iteration_name = combination
                            modality_timeseries, modality_timeseries_noiseless = experiment.run_simulation(
                                name="", save=False, modality=modality_name
                            )
                            mod_comb = (
                                vsmpl_id
                                + "_"
                                + str(mod_n)
                                + "_"
                                + str(mod_pars_number)
                                + "_"
                                + str(mod_acq_number)
                            )
                            mod_parameters = copy.copy(vsampl_pars[vsmpl_id])
                            mod_parameters.append(modality_name)
                            pxsize = (
                                experiment.imager.modalities[modality_name][
                                    "detector"
                                ]["pixelsize"]
                                * 1000
                            )
                            mod_params_copy = copy.deepcopy(mod_pars)
                            mod_params_copy["pixelsize"] = pxsize
                            mod_parameters.append(mod_params_copy)
                            mod_parameters.append(acq_pars)
                            if mod_comb not in mod_params.keys():
                                mod_params[mod_comb] = mod_parameters
                                mod_outputs[mod_comb] = []
                            if mod_comb not in mod_outputs_masks.keys():
                                mod_outputs_masks[mod_comb] = []
                            mod_outputs[mod_comb].append(
                                modality_timeseries[modality_name][channel]
                            )
                            #mask = modality_timeseries[modality_name][0] > mod_threshold
                            mask = vsample_mask_per_mod > 0
                            mod_outputs_masks[mod_comb].append(
                                mask
                            )
                            mod_parameters = None
                    mod_n += 1
    return experiment, mod_outputs, mod_params, pixelsizes, mod_outputs_masks


def generate_global_reference_sample(
    experiment: ExperimentParametrisation = None,
    structure=None,
    probe=None,
    probe_parameters=None,
    virtual_sample=None,
    structure_is_path=False,
    **kwargs
):
    """
    Generate a global reference virtual sample.

    Parameters
    ----------
    experiment : ExperimentParametrisation, optional
        Experiment object to use. If None, a new one is created.
    structure : str, optional
        Structure ID.
    probe : str, optional
        Probe name.
    probe_parameters : dict, optional
        Probe parameters.
    virtual_sample : dict, optional
        Virtual sample parameters.

    Returns
    -------
    reference_vsample : dict
        The generated reference virtual sample.
    refernece_parameters : list
        List of parameters used for the reference.
    """
    # empty lists, but fill up with default
    pck_dir = os.path.dirname(os.path.abspath(vlab4mic.__file__))
    local_dir = os.path.join(pck_dir, "configs")
    default_reference_probe = "NHS_ester"
    default_reference_fluorophore = "AF647"
    if experiment is None:
        experiment = ExperimentParametrisation()
        experiment.clear_experiment()
    if probe is None:
        probe = default_reference_probe
    if probe_parameters is None:
        probe_filepath = os.path.join(
            local_dir, "probes", default_reference_probe + ".yaml"
        )
        probe_parameters = load_yaml(probe_filepath)
        probe_parameters["fluorophore_id"] = default_reference_fluorophore
    else:
        probe_parameters["fluorophore_id"] = default_reference_fluorophore
    if structure_is_path:
        experiment.select_structure(
            structure_path=structure, structure_id="global_reference"
        )
    else:
        experiment.structure_id = structure
        experiment._build_structure()
    experiment.add_probe(probe_template=probe, **probe_parameters, **kwargs)
    # experiment.structure_label = probe
    # experiment.probe_parameters[probe] = probe_parameters
    experiment._build_particle(keep=True)
    experiment.set_virtualsample_params(**kwargs)
    # combination += str(vsample_n)
    probe_parameters = experiment.probe_parameters[probe]
    refernece_parameters = [structure, probe, probe_parameters, virtual_sample]
    reference_vsample = experiment._build_coordinate_field(
        keep=False, use_self_particle=True
    )
    return reference_vsample, refernece_parameters


def generate_global_reference_modality(
    experiment: ExperimentParametrisation = None,
    reference_vsample=None,
    reference_vsample_params=None,
    modality=None,
    modality_acquisition=None,
    mod_threshold = None
):
    """
    Generate a global reference image for a given virtual sample and modality.

    Parameters
    ----------
    experiment : ExperimentParametrisation, optional
        Experiment object to use. If None, a new one is created.
    reference_vsample : dict, optional
        Reference virtual sample.
    reference_vsample_params : dict, optional
        Parameters for the reference virtual sample.
    modality : str, optional
        Modality name.
    modality_acquisition : dict, optional
        Acquisition parameters for the modality.

    Returns
    -------
    reference_output : numpy.ndarray
        The simulated reference image.
    reference_parameters : dict
        Parameters used for the reference image.
    """
    if experiment is None:
        experiment = ExperimentParametrisation()
        experiment.clear_experiment()
    if modality is None:
        modality = "Reference"
        # modality_acquisition = None
    if modality_acquisition is None:
        experiment.add_modality(modality)

        # experiment.selected_mods[modality] = (
        #    configuration_format.format_modality_acquisition_params()
        # )
    else:
        experiment.add_modality(modality)
        experiment.set_modality_acq(modality, **modality_acquisition)
        # experiment.selected_mods[modality] = modality_acquisition
    experiment._build_imager(use_local_field=False)
    print(experiment.generators_status)
    experiment.imager.import_field(**reference_vsample)
    reference_parameters = dict()
    reference_parameters["Vector"] = [
        reference_vsample_params,
        modality,
        modality_acquisition,
    ]
    reference_output, reference_output_noiseless = experiment.run_simulation(name="", save=False)
    imager_scale = experiment.imager.roi_params["scale"]
    scalefactor = np.ceil(
        imager_scale / 1e-9
    )  # resulting pixel size in nanometers
    reference_parameters["ref_pixelsize"] = (
        experiment.imager.modalities[modality]["detector"]["pixelsize"]
        * scalefactor
    )
    #if mod_threshold is None:
    #    mod_threshold = 1
    reference_output_mask = experiment.imager.generate_modality_mask(modality="Reference")
    return reference_output[modality]["ch0"], reference_parameters, reference_output_mask


def analyse_sweep_single_reference(
    img_outputs,
    img_outputs_masks,
    img_params,
    reference_image,
    reference_image_mask,
    reference_params,
    zoom_in=0,
    metrics: dict = None,
    #custom_metrics = None,
    **kwargs,
):
    """
    Analyse a sweep of images against a single reference image using specified metrics.

    Parameters
    ----------
    img_outputs : dict
        Dictionary of simulated image outputs.
    img_params : dict
        Dictionary of image parameters.
    reference_image : numpy.ndarray
        Reference image.
    reference_params : dict
        Parameters for the reference image.
    zoom_in : float, optional
        Zoom factor for analysis. Defaults to 0.
    metrics_list : list of str, optional
        List of metric names to compute. Defaults to ["ssim"].
    **kwargs
        Additional keyword arguments.

    Returns
    -------
    measurement_vectors : list
        List of measurement results for each image.
    inputs : dict
        Dictionary of input images and used references.
    metrics_list : list
        List of metric names used.
    """
    measurement_vectors = []
    # ref_pixelsize = analysis_case_params["ref_pixelsize"]
    inputs = dict()
    for params_id in img_params.keys():
        inputs[params_id] = dict()
        rep_number = 0
        mod_name = img_params[params_id][5]  # 5th item corresponds to Modality
        modality_pixelsize = img_params[params_id][6]["pixelsize"]
        for img_r, img_mask in zip(img_outputs[params_id], img_outputs_masks[params_id]):
            im1 = img_r[0]
            #im1_mask = img_mask
            #im_ref = reference_image
            similarity_vector = []
            metrics_names_list = []
            for metric_name, metric in metrics.items():
                #metrics_names_list.append(metric_name)
                similarity_vector.append(metric(
                        reference_image = reference_image, 
                        reference_image_pixelsize_nm = reference_params["ref_pixelsize"],
                        reference_image_mask = reference_image_mask,
                        simulated_image = im1,
                        simulated_image_pixelsize_nm = modality_pixelsize,
                        simulated_image_mask = img_mask,
                    )
                )
                #rep_measurement, ref_used, qry_used, masks_used = metrics.img_compare(
                #    ref = im_ref,
                #    ref_mask=reference_image_mask,
                #    query=im1,
                #    query_mask=im1_mask,
                #    modality_pixelsize=modality_pixelsize,
                #    ref_pixelsize=reference_params["ref_pixelsize"],
                #    force_match=True,
                #    zoom_in=zoom_in,
                #    metrics = metrics
                #)
            # each image as its ID, a replica number and the metrics associated to it
            replicaID_repN_metrics = list([params_id, rep_number]) + list([*similarity_vector])
            measurement_vectors.append(replicaID_repN_metrics)
                # for methods that require image resizing
                #inputs[params_id][rep_number] = [qry_used, im1, ref_used, masks_used]
            rep_number += 1
    return measurement_vectors, inputs, metrics_names_list


def measurements_dataframe(
    measurement_vectors,
    probe_parameters=None,
    p_structural_integrity=None,
    mod_names=None,
    sample_params=None,
    mod_params=None,
    mod_acq=None,
    metric_names=None,
):
    """
    Create a pandas DataFrame from measurement vectors and parameter dictionaries.

    Parameters
    ----------
    measurement_vectors : list
        List of measurement results.
    probe_parameters : dict, optional
        Dictionary of probe parameter sets.
    p_structural_integrity : dict, optional
        Dictionary of structural integrity parameter sets.
    mod_names : list, optional
        List of modality names.
    sample_params : dict, optional
        Dictionary of sample parameter sets.
    mod_params : dict, optional
        Dictionary of modality parameter sets.
    mod_acq : dict, optional
        Dictionary of modality acquisition parameter sets.
    metric_names : list, optional
        List of metric names.

    Returns
    -------
    data_frame : pandas.DataFrame
        DataFrame with basic combination and replica info.
    df_combined : pandas.DataFrame
        DataFrame with expanded parameter and metric columns.
    """
    measurement_array = np.array(measurement_vectors)
    nrows = len(measurement_array[:, 1])
    ids = [i.split("_") for i in measurement_array[:, 0]]
    ids_array = np.array(ids)
    data_frame = pd.DataFrame(
        data={
            "Combination_id": measurement_array[:, 0],
            "probe_n": ids_array[:, 0],
            "probe_param_n": ids_array[:, 1],
            "structural_integrity_n": ids_array[:, 2],
            "vsample": ids_array[:, 3],
            "modality": ids_array[:, 4],
            "modality_parameters": ids_array[:, 5],
            "modality_acqusition": ids_array[:, 6],
            "Replica": measurement_array[:, 1],
        }
    )
    df_combined = data_frame
    #
    nmetrics = len(metric_names)
    metrics_dictionary = dict()
    for metric_number in range(nmetrics):
        # first two indices are ID and replica number, then one value per metric
        metricvector = measurement_array[:, 2 + metric_number]
        metrics_dictionary[metric_names[metric_number]] = np.array(
            metricvector, dtype=np.float64
        )
    metrics_df = pd.DataFrame(metrics_dictionary)
    df_combined = df_combined.join(metrics_df)
    if probe_parameters:
        probe_param_names = probe_parameters[0].keys()
        tmp_df1 = dict()
        for column_name in probe_param_names:
            tmp_df1[column_name] = []
        for i in range(nrows):
            probe_par_comb_id = int(data_frame.iloc[i]["probe_param_n"])
            for column_name in probe_param_names:
                tmp_df1[column_name].append(
                    probe_parameters[probe_par_comb_id][column_name]
                )
        tmp1 = pd.DataFrame(tmp_df1)
        df_combined = df_combined.join(tmp1)
    if p_structural_integrity:
        structural_integrity_param_names = p_structural_integrity[0].keys()
        tmp_df2 = dict()
        for column_name in structural_integrity_param_names:
            tmp_df2[column_name] = []
        for i in range(nrows):
            structural_integrity_par_comb_id = int(data_frame.iloc[i]["structural_integrity_n"])
            for column_name in structural_integrity_param_names:
                tmp_df2[column_name].append(
                    p_structural_integrity[structural_integrity_par_comb_id][column_name]
                )
        tmp2 = pd.DataFrame(tmp_df2)
        df_combined = df_combined.join(tmp2)
    if sample_params:
        sample_param_names = sample_params[0].keys()
        tmp_df3 = dict()
        for column_name in sample_param_names:
            tmp_df3[column_name] = []
        for i in range(nrows):
            sample_par_comb_id = int(data_frame.iloc[i]["vsample"])
            for column_name in sample_param_names:
                tmp_df3[column_name].append(
                    sample_params[sample_par_comb_id][column_name]
                )
        tmp3 = pd.DataFrame(tmp_df3)
        df_combined = df_combined.join(tmp3)
    if mod_params:
        param_names = mod_params[0].keys()
        tmp_df4 = dict()
        for column_name in param_names:
            tmp_df4[column_name] = []
        for i in range(nrows):
            acq_par_comb_id = int(data_frame.iloc[i]["modality_parameters"])
            for column_name in param_names:
                tmp_df4[column_name].append(
                    mod_params[acq_par_comb_id][column_name]
                )
        tmp4 = pd.DataFrame(tmp_df4)
        df_combined = df_combined.join(tmp4)
    if mod_acq:
        param_names = mod_acq[0].keys()
        tmp_df5 = dict()
        for column_name in param_names:
            tmp_df5[column_name] = []
        for i in range(nrows):
            acq_par_comb_id = int(data_frame.iloc[i]["modality_acqusition"])
            for column_name in param_names:
                tmp_df5[column_name].append(
                    mod_acq[acq_par_comb_id][column_name]
                )
        tmp5 = pd.DataFrame(tmp_df5)
        df_combined = df_combined.join(tmp5)
    if mod_names:
        tmp_df6 = {}
        tmp_df6["modality_name"] = []
        mod_names_id = {}
        for m in range(len(mod_names)):
            mod_names_id[m] = mod_names[m]
        for i in range(nrows):
            acq_par_comb_id = int(data_frame.iloc[i]["modality"])
            tmp_df6["modality_name"].append(mod_names_id[acq_par_comb_id])
        tmp6 = pd.DataFrame(tmp_df6)
        df_combined = df_combined.join(tmp6)

    return data_frame, df_combined


def create_param_combinations(**kwargs):
    """
    Generate all combinations of parameter values.

    Parameters
    ----------
    **kwargs
        Parameter names and their possible values (lists).

    Returns
    -------
    combinations_dict : dict
        Dictionary mapping combination index to parameter dictionary.
    """
    if kwargs:
        # Generate all combinations using itertools.product
        combinations = list(itertools.product(*kwargs.values()))
        # Create a new dictionary with unique integers as keys
        # Each value in the dictionary will be another dictionary where the keys are the parameter names
        combinations_dict = {
            i: {key: value for key, value in zip(kwargs.keys(), combination)}
            for i, combination in enumerate(combinations)
        }
        return combinations_dict


def pivot_dataframes_byCategory(
    dataframe, category_name, param1, param2, metric_name, pre_filter_dt=False, filter_dictionary:dict = None, **kwargs
):
    """
    Pivot a DataFrame by category and two parameters, summarizing a metric.

    Parameters
    ----------
    dataframe : pandas.DataFrame
        DataFrame to pivot.
    category_name : str
        Name of the category column.
    param1 : str
        First parameter for pivot index.
    param2 : str
        Second parameter for pivot columns.
    metric_name : str
        Name of the metric to summarize.
    **kwargs
        Additional keyword arguments.

    Returns
    -------
    df_categories : dict
        Dictionary of pivoted DataFrames for each category.
    titles : dict
        Dictionary with category, param1, and param2 names.
    """
    df_categories = dict()
    dataframe_copy = copy.deepcopy(dataframe)
    if pre_filter_dt and filter_dictionary is not None:
        for parameter_name, parameter_value in filter_dictionary.items():
            dataframe_copy = dataframe_copy[dataframe_copy[parameter_name] == parameter_value]
        #print(dataframe_copy)
    for category, group in dataframe_copy.groupby(category_name):
        summarised_group = None
        summarised_group = (
            group.groupby([param1, param2])
            .agg(
                Mean_Value=(metric_name, "mean"),
                Std_Dev=(metric_name, "std"),
                Replicas_Count=("Replica", "count"),
            )
            .reset_index()
        )
        # get mean and std across parameter combinations of axes_param_names
        condition_mean_pivot = summarised_group.pivot(
            index=param1, columns=param2, values="Mean_Value"
        ).round(4)
        condition_sd_pivot = summarised_group.pivot(
            index=param1, columns=param2, values="Std_Dev"
        ).round(4)
        df_categories[str(category)] = [
            condition_mean_pivot,
            condition_sd_pivot,
        ]
    titles = dict(category=category_name, param1=param1, param2=param2)
    return df_categories, titles


def probe_parameters_sweep(
    probe_target_type: list[str] = None,
    probe_target_value: list[str] = None,
    probe_distance_to_epitope: float = None,
    probe_model: list[str] = None,
    probe_fluorophore: str = None,
    probe_paratope: str = None,
    probe_conjugation_target_info=None,
    probe_DoL: list[float] = None,
    probe_seconday_epitope=None,
    probe_wobbling=None,
    labelling_efficiency: list[float] = None,
):
    """
    Generate combinations of probe parameters for a sweep.

    Parameters
    ----------
    probe_target_type : list of str, optional
    probe_target_value : list of str, optional
    probe_distance_to_epitope : float, optional
    probe_model : list of str, optional
    probe_fluorophore : str, optional
    probe_paratope : str, optional
    probe_conjugation_target_info : any, optional
    probe_DoL : list of float, optional
    probe_seconday_epitope : any, optional
    probe_wobbling : any, optional
    labelling_efficiency : list of float, optional

    Returns
    -------
    probe_parameters : dict or None
        Dictionary of parameter combinations, or None if no parameters.
    """
    local_params = locals()
    probe_parameters_vectors = {}
    for par, value in local_params.items():
        if value is not None and type(value) is list:
            if len(value) == 1:
                probe_parameters_vectors[par] = value
            else:
                if isinstance(value[0], (str, bool)):
                    probe_parameters_vectors[par] = value
                else:
                    sequence = np.linspace(value[0], value[1], value[2])
                    probe_parameters_vectors[par] = sequence
    probe_parameters = None
    if bool(probe_parameters_vectors):
        probe_parameters = create_param_combinations(
            **probe_parameters_vectors
        )
    return probe_parameters


def virtual_sample_parameters_sweep(
    virtual_sample_template: str = None,
    sample_dimensions: list[float] = None,
    number_of_particles: int = None,
    particle_positions: list[np.array] = None,
    random_orientations=False,
    random_placing=False,
):
    """
    Generate combinations of virtual sample parameters for a sweep.

    Parameters
    ----------
    virtual_sample_template : str, optional
    sample_dimensions : list of float, optional
    number_of_particles : int, optional
    particle_positions : list of numpy.array, optional
    random_orientations : bool, optional
    random_placing : bool, optional

    Returns
    -------
    field_parameters : dict or None
        Dictionary of parameter combinations, or None if no parameters.
    """
    local_params = locals()
    field_parameters_vectors = {}
    for par, value in local_params.items():
        if value is not None and type(value) is list:
            if len(value) == 1:
                field_parameters_vectors[par] = value
            else:
                if isinstance(value[0], (str, bool)):
                    field_parameters_vectors[par] = value
                else:
                    sequence = np.linspace(value[0], value[1], value[2])
                    field_parameters_vectors[par] = sequence
    field_parameters = None
    if bool(field_parameters_vectors):
        field_parameters = create_param_combinations(
            **field_parameters_vectors
        )
    return field_parameters


def structural_integrity_parameters_sweep(
    structural_integrity_small_cluster: float = None,
    structural_integrity_large_cluster: float = None,
    structural_integrity: float = None,
):
    """
    Generate combinations of structural integrity parameters for a sweep.

    Parameters
    ----------
    structural_integrity_small_cluster : float, optional
    structural_integrity_large_cluster : float, optional
    structural_integrity : float, optional

    Returns
    -------
    structural_integrity_parameters : dict or None
        Dictionary of parameter combinations, or None if no parameters.
    """
    local_params = locals()
    structural_integrity_parameters_vectors = {}
    for par, value in local_params.items():
        if value is not None and type(value) is list:
            if len(value) == 1:
                structural_integrity_parameters_vectors[par] = value
            else:
                if isinstance(value[0], (str, bool)):
                    structural_integrity_parameters_vectors[par] = value
                else:
                    sequence = np.linspace(value[0], value[1], value[2])
                    structural_integrity_parameters_vectors[par] = sequence
    structural_integrity_parameters = None
    if bool(structural_integrity_parameters_vectors):
        structural_integrity_parameters = create_param_combinations(
            **structural_integrity_parameters_vectors
        )
    return structural_integrity_parameters


def modality_parameters_sweep(
    # modality_name: str = None,
    pixelsize_nm: float = None,
    lateral_resolution_nm: float = None,
    axial_resolution_nm: float = None,
    psf_voxel_nm: int = None,
):
    """
    Generate combinations of modality parameters for a sweep.

    Parameters
    ----------
    pixelsize_nm : float, optional
    lateral_resolution_nm : float, optional
    axial_resolution_nm : float, optional
    psf_voxel_nm : int, optional

    Returns
    -------
    modality_parameters : dict or None
        Dictionary of parameter combinations, or None if no parameters.
    """
    local_params = locals()
    modality_parameters_vectors = {}
    for par, value in local_params.items():
        if value is not None and type(value) is list:
            if len(value) == 1:
                modality_parameters_vectors[par] = value
            else:
                if isinstance(value[0], (str, bool)):
                    modality_parameters_vectors[par] = value
                else:
                    sequence = np.linspace(value[0], value[1], value[2])
                    modality_parameters_vectors[par] = sequence
    modality_parameters = None
    if bool(modality_parameters_vectors):
        modality_parameters = create_param_combinations(
            **modality_parameters_vectors
        )
    return modality_parameters


def acquisition_parameters_sweep(
    exp_time: str = None,
    noise: float = None,
    nframes: float = None,
    channels: float = None,
):
    """
    Generate combinations of acquisition parameters for a sweep.

    Parameters
    ----------
    exp_time : str, optional
    noise : float, optional
    nframes : float, optional
    channels : float, optional

    Returns
    -------
    acq_parameters : dict or None
        Dictionary of parameter combinations, or None if no parameters.
    """
    local_params = locals()
    acq_parameters_vectors = {}
    for par, value in local_params.items():
        if value is not None and type(value) is list:
            if len(value) == 1:
                acq_parameters_vectors[par] = value
            else:
                if isinstance(value[0], (str, bool)):
                    acq_parameters_vectors[par] = value
                else:
                    sequence = np.linspace(value[0], value[1], value[2])
                    acq_parameters_vectors[par] = sequence
    acq_parameters = None
    if bool(acq_parameters_vectors):
        acq_parameters = create_param_combinations(**acq_parameters_vectors)
    return acq_parameters



from vlab4mic.generate import labels
from vlab4mic.workflows import probe_model
from vlab4mic.utils.io.yaml_functions import load_yaml
import os


def test_createlabel_from_config(configuration_directory):
    labelcongf_filename = "NPC_Nup96_Cterminal_direct.yaml"
    local_path_congfig = os.path.join(
        configuration_directory, "probes", labelcongf_filename
    )
    labelling_efficiency = 0.9
    probe_params = load_yaml(local_path_congfig)
    label_object, label_params = labels.construct_label(
        label_config_dictionary=probe_params, lab_eff=labelling_efficiency
    )


def test_probe_model(configuration_directory):
    label_file = "Antibody" + ".yaml"
    local_path_congfig = os.path.join(
        configuration_directory, "probes", label_file
    )
    probe_params = load_yaml(local_path_congfig)
    probe, probe_emitters, anchor, ab_ref, probe_epitope = probe_model(
        **probe_params, config_dir=configuration_directory
    )
    assert anchor.shape == (3,)
    assert ab_ref.shape == (3,)



from .generate.molecular_structure import build_structure_cif
from .utils.io.yaml_functions import load_yaml
from .generate.molecular_structure import MolecularReplicates
from .generate import labelled_instance as labinstance
from .generate.labels import construct_label
from .generate.coordinates_field import create_min_field
from .generate.imaging import Imager
from .utils.data_format.configuration_format import (
    compile_modality_parameters,
    format_modality_acquisition_params,
)
from .download import verify_structure
import os
import copy
import itertools
import matplotlib.pyplot as plt
from IPython.utils import io
from tqdm import tqdm
import numpy as np
from scipy.spatial.distance import pdist
from .utils.transform.datatype import truncate
from .utils.data_format.structural_format import label_builder_format
from pathlib import Path

def load_structure(structure_id: str = None, config_dir=None, structure_path=None, structure_format="CIF"):
    """
    Initialise a BioPython object for the PDB/CIF ID and retrieve available information about specific labelling.

    This function requires your configuration directory to contain a structure configuration file (yaml).
    It automatically downloads the CIF file of the structure if it does not exist locally.

    Parameters
    ----------
    structure_id : str, optional
        4-letter ID of structure.
    config_dir : str, optional
        Absolute path for configuration files.

    Returns
    -------
    structure : Molecularstructure
        The loaded structure object.
    structure_params : dict
        Structure parameters loaded from configuration.
    """
    if config_dir is not None:
        if structure_path is not None:
            print("Loading structure from local file")
            structure = build_structure_cif(
                cif_file=structure_path, struct_title="", cif_id=structure_id, format_type=structure_format
            )
            structure_params = dict(
                model = {"ID": structure_id, "format": structure_format, "title": "", "database": ""},
            )
            print("Structure Loaded!")
            return structure, structure_params
        else:
            print("Loading structure.")
            structure_dir = os.path.join(config_dir, "structures")
            config_file = structure_id + ".yaml"
            structure_configuration = os.path.join(structure_dir, config_file)
            structure_configuration_check = Path(structure_configuration)
            if structure_configuration_check.exists():
                structure_params = load_yaml(structure_configuration)
            else:
                template = os.path.join(structure_dir, "_template_.yaml")
                structure_params = load_yaml(template)
            # get CIF path
            cif_file = verify_structure(structure_id, structure_dir)
            # build structure
            title = ""
            if "title" in structure_params["model"]:
                title = structure_params["model"]["title"]
            structure = build_structure_cif(
                cif_file=cif_file,
                struct_title=title,
                cif_id=structure_id,
            )
            print("Structure Loaded!")
            if "labels" not in structure_params.keys():
                print("Structure has no predefined labels")
                structure_params["labels"] = list(["<None>"])
            return structure, structure_params
    else:
        print("No configuration directory exists")


def probe_model(
    model, binding, conjugation_sites, config_dir, probe_name="probe", **kwargs
):
    """
    Generate probe model and anchor points for a given structure and binding configuration.

    Parameters
    ----------
    model : dict
        Model information dictionary.
    binding : dict
        Binding configuration dictionary.
    conjugation_sites : dict
        Conjugation site information.
    config_dir : str
        Path to configuration directory.
    probe_name : str, optional
        Name for the probe (default "probe").
    **kwargs
        Additional keyword arguments, e.g., epitope.

    Returns
    -------
    structural_model : MolecularReplicates
        The structure with generated targets.
    target_sites : numpy.ndarray
        Coordinates of target sites.
    anchor_point : numpy.ndarray
        Anchor point for the probe.
    direction_point : numpy.ndarray
        Direction point for the probe.
    probe_epitope : dict
        Dictionary with epitope coordinates and normals.
    """
    anchor_point = None
    probe_epitope = dict(coordinates=None, normals=None)
    structural_model, structure_model_prams = load_structure(model["ID"], config_dir)
    # generate conjugation sites
    # print(conjugation_sites["target"]["value"])
    structural_model.gen_Targets(
        target_name=probe_name,
        target_type=conjugation_sites["target"]["type"],
        target_value=conjugation_sites["target"]["value"],
        **kwargs,
    )
    target_sites = structural_model.label_targets[probe_name]["coordinates"]
    # calculate axis for antibody by getting center of mass and paratope sequence
    if binding["paratope"]:
        structural_model.gen_Targets(
            target_name="paratope",
            target_type="Sequence",
            target_value=binding["paratope"],
            **kwargs,
        )
        paratopes = structural_model.label_targets["paratope"]["coordinates"]
        if paratopes.shape[0] > 1:
            print("more than one paratope sites recovered. Calculating average")
            paratope_site = np.mean(
                structural_model.label_targets["paratope"]["coordinates"], axis=0
            )
        else:
            paratope_site = structural_model.label_targets["paratope"]["coordinates"]
    direction_point = structural_model.assembly_refpt
    # extend the anchor from paratope
    diff = np.array([0, 0, 0])
    if binding["distance"]["to_target"] is not None and binding["paratope"]:
        difference = paratope_site - direction_point
        unit_vector = difference / np.linalg.norm(difference)
        diff = unit_vector * binding["distance"]["to_target"]
    else:
        binding["distance"]["to_target"] = 0
    print(f"dif:{diff}")
    anchor_point = paratope_site + diff
    # generate a site for binding
    if "epitope" in kwargs.keys():
        print("generating epitope")
        try:
            #print(kwargs["epitope"])
            structural_model.gen_Targets(
                target_name="epitope",
                target_type=kwargs["epitope"]["target"]["type"],
                target_value=kwargs["epitope"]["target"]["value"],
                **kwargs,
            )
            epitope_sites = structural_model.label_targets["epitope"]["coordinates"]
            if epitope_sites.shape[0] > 1:
                print("more than one paratope sites recovered. Calculating average")
                epitope_sites = np.mean(
                    structural_model.label_targets["epitope"]["coordinates"], axis=0
                )
            probe_epitope["coordinates"] = epitope_sites
            print(f"epitope site: {probe_epitope}")
        except:
            print("No epitope generated")
            print(kwargs["epitope"])

    # print(structural_model.label_targets)
    

    return structural_model, target_sites, np.squeeze(anchor_point, axis=None), np.squeeze(direction_point, axis=None), probe_epitope


def particle_from_structure(
    structure: MolecularReplicates, labels=list, config_dir=None
):
    """
    Create a labelled particle from a structure and label definitions.

    Parameters
    ----------
    structure : MolecularReplicates
        Object representing the parsed CIF structure.
    labels : list of dict
        List of label dictionaries, each containing label ID and fluorophore ID.
    config_dir : str, optional
        Path to configuration directory.

    Returns
    -------
    particle : LabelledInstance
        The created labelled particle.
    label_params_list : list of dict
        List of label parameter dictionaries.
    """
    if config_dir is not None:
        label_params_list = []
        label_config_dir = os.path.join(config_dir, "probes")
        for label in labels:
            #label_name = label["label_id"] + ".yaml"
            #label_config_path = os.path.join(label_config_dir, label_name)
            label_object, label_params = construct_label(
                label_config_dictionary=label
            )
            # get model for antibody and add to label params
            if label_object.model:
                print("Generating conjugation sites")
                (
                    probe,
                    probe_emitter_sites,
                    anchor_point,
                    direction_point,
                    probe_epitope,
                ) = probe_model(
                    model=label_object.model,
                    binding=label_object.binding,
                    conjugation_sites=label_object.conjugation,
                    epitope=label_object.epitope,
                    config_dir=config_dir,
                )
                if anchor_point.shape == (3,):
                    print("setting new axis")
                    print(anchor_point, direction_point)
                    direction_vector = direction_point - anchor_point
                    label_object.set_axis(pivot=anchor_point, direction=direction_vector)
                    print(f'label object axis: {label_object.params["axis"]}')
                if (
                    probe_epitope["coordinates"] is not None
                    and label_params["as_linker"]
                ):
                    # as_linker is True when this probe has a secondary linker
                    print("Generating linker from epitope site")
                    label_object.set_emitters(probe_epitope["coordinates"])
                else:
                    print(probe_epitope["coordinates"], label_params["as_linker"])
                    print("Setting emitters from probe conjugation sites")
                    label_object.set_emitters(probe_emitter_sites)
                label_params["coordinates"] = label_object.gen_labeling_entity()
                label_params["structural_atoms"] = copy.deepcopy(probe.assembly_atoms)
                print(label_params["coordinates"])
            
            structure.add_label(label_object)
            if label_params["binding"]["distance"]["to_target"] is not None:
                print("Label is indirect label")
                structure.assign_normals2targets()  # default is with scaling
            else:
                print("Label is direct")
            if label_params["binding"]["distance"]["between_targets"] == "estimate":
                distances = pdist(label_params["coordinates"])
                label_params["binding"]["distance"]["between_targets"] = np.max(distances)
            label_params_list.append(label_params)
        inst_builder = structure.create_instance_builder()
        particle = labinstance.create_particle(
            source_builder=inst_builder, label_params_list=label_params_list
        )
        return particle, label_params_list


def field_from_particle(
    particle: labinstance.LabeledInstance, field_config: str = None, **kwargs
):
    """
    Create a particle field from an input particle object.

    A minimal field is initialised; by default it defines a single particle at the middle of a square area of 1x1 micrometers.
    If a field configuration file is provided, the initialised field is adjusted from this configuration file.

    Parameters
    ----------
    particle : LabeledInstance
        Particle object to place in the field.
    field_config : dict or str, optional
        Field configuration dictionary or file path.
    **kwargs
        Additional keyword arguments for field creation.

    Returns
    -------
    exported_field : dict
        Exported field dictionary.
    coordinates_field : CoordinatesField
        The coordinates field object.
    """
    if field_config is not None:
        print("Creating field from parameter files")
        # coordinates_field.init_from_file(field_config)
        ## if not file, use as dictionary
        coordinates_field = create_min_field(**field_config, **kwargs)
        # coordinates_field.init_from_file(field_config)
    else:
        coordinates_field = create_min_field(**kwargs)
    coordinates_field.create_molecules_from_InstanceObject(particle)
    coordinates_field.construct_static_field()
    return coordinates_field.export_field(), coordinates_field


def create_imaging_system(
    exported_field=None, modalities_id_list: list = None, config_dir=None, mod_params=dict(), fluorophore_parameters=None, **kwargs
):
    """
    Create an imaging system object.

    Reads the configuration files for the specified imaging modalities in modalities_id_list and initialises the imager object.

    Parameters
    ----------
    exported_field : dict, optional
        Output dictionary from the export_field method of the coordinate field.
    modalities_id_list : list of str, optional
        List of modality IDs. The IDs must match the name of their configuration file.
    config_dir : str, optional
        Path to configuration directory.
    mod_params : dict, optional
        Dictionary of modality parameters.
    **kwargs
        Additional keyword arguments.

    Returns
    -------
    image_generator : Imager
        Instance of Imaging class initialised with the input modalities.
    modality_parameters : list of dict
        List of modality parameter dictionaries.
    """
    if config_dir is not None:
        if exported_field is None:
            # minimal field
            coordinates_field = create_min_field()
            exported_field = coordinates_field.export_field()
            image_generator = Imager()
            image_generator.import_field(**exported_field)
        else:
            image_generator = Imager()
            image_generator.import_field(**exported_field)
        fluo_emission = dict()
        for fluo in exported_field["field_emitters"].keys():
            #fluo_dir = os.path.join(config_dir, "fluorophores", fluo)
            #fluopath = fluo_dir + ".yaml"
            # image_generator.set_fluorophores_from_file(fluopath)
            #try:
            #    fluo_params = load_yaml(fluopath)
            #except:
            fluo_params = fluorophore_parameters[fluo]
            image_generator.set_fluorophores_params(
                identifier=fluo,
                photon_yield=fluo_params["emission"]["photon_yield"],
                emission=fluo_params["emission"]["type"],
                blinking_rates=fluo_params["blinking_rates"],
            )
            fluo_emission[fluo] = fluo_params["emission"]["type"]
        modality_parameters = []
        for mod in modalities_id_list:
            if mod in mod_params.keys():
                image_generator.set_imaging_modality(**mod_params[mod])
                modality_parameters.append(mod_params[mod])
            else:
                modality = compile_modality_parameters(mod, config_dir, fluo_emission)
                modality_parameters.append(modality)
                image_generator.set_imaging_modality(**modality)
        return image_generator, modality_parameters
        


# generate several modalities results
def generate_multi_imaging_modalities(
    image_generator,
    experiment_name="multi_imaging_modalities",
    savingdir=None,
    acquisition_param: dict = None,
    write=False,
    **kwargs,
):
    """
    Generate imaging outputs for each modality from the image generator.

    Parameters
    ----------
    image_generator : Imager
        Instance of Imaging class.
    experiment_name : str, optional
        Name to add to each output file. Default is "multi_imaging_modalities".
    savingdir : str, optional
        Output directory for saving images.
    acquisition_param : dict, optional
        Dictionary containing acquisition parameters for each modality.
    write : bool, optional
        If True, all output images will be written to savingdir.
    **kwargs
        Additional keyword arguments.

    Returns
    -------
    outputs : dict
        Dictionary of image outputs per modality.
    """
    # kwargs will contain as keys the names of the modalities,
    # which shall correspond to the modality name
    # in the yaml file. each modality will have specific parameters for aquisitions
    # there must be a default imaging parameter for all, like 10 frames for each
    image_generator.set_experiment_name(experiment_name)
    outputs = dict()
    outputs_noiseless = dict()
    if acquisition_param is None:
        print("No acquisition parameters defined. Using default on all modalities")
        for mod in image_generator.modalities.keys():
            # should verify that the path exist
            if savingdir is not None:
                savingdir = savingdir + os.sep
                image_generator.set_writing_directory(savingdir)
            acq_params = format_modality_acquisition_params(save=write)
            timeseries, calibration_beads, timeseries_noiseless, calibration_beads_noiseless  = image_generator.generate_imaging(
                modality=mod, **acq_params
            )
            outputs[mod] = timeseries
            outputs_noiseless[mod] = timeseries_noiseless
    else:
        acquisition_parameters = copy.copy(acquisition_param)
        for mod, acq_param in acquisition_parameters.items():
            #print(mod, acq_params)
            if acq_param is None:
                acq_params = format_modality_acquisition_params()
            else:
                acq_params = format_modality_acquisition_params(**acq_param)
            acq_params["save"] = write
            if savingdir is not None:
                savingdir = savingdir + os.sep
                image_generator.set_writing_directory(savingdir)
                timeseries, calibration_beads, timeseries_noiseless, calibration_beads_noiseless = image_generator.generate_imaging(
                    modality=mod, **acq_params
                )
                outputs[mod] = timeseries
                outputs_noiseless[mod] = timeseries_noiseless
    return outputs, outputs_noiseless



import matplotlib.pyplot as plt
import mpl_toolkits.axes_grid1 as axes_grid1
import numpy as np
import copy
from scipy.ndimage import rotate


def add_ax_scatter(plotobj, trgt_dictionary, fraction=1):
    if trgt_dictionary["coordinates"].shape[0] == 0:
        print(f"No emitters to show.")
    else:
        if fraction == 1:
            plotobj.scatter(
                trgt_dictionary["coordinates"][:, 0],
                trgt_dictionary["coordinates"][:, 1],
                trgt_dictionary["coordinates"][:, 2],
                c=trgt_dictionary["plotcolour"],
                label=trgt_dictionary["label_name"],
                s=trgt_dictionary["plotsize"],
                alpha=trgt_dictionary["plotalpha"],
                marker=trgt_dictionary["plotmarker"],
                depthshade=True,
            )
        else:
            n = ((trgt_dictionary["coordinates"]).shape)[0]
            print(f"Showing {n*fraction} atoms for {trgt_dictionary['label_name']}")
            ids = np.random.choice(np.arange(0, n), int(n * fraction), replace=False)
            subset = trgt_dictionary["coordinates"][ids, :]
            plotobj.scatter(
                subset[:, 0],
                subset[:, 1],
                subset[:, 2],
                c=trgt_dictionary["plotcolour"],
                label=trgt_dictionary["label_name"],
                s=trgt_dictionary["plotsize"],
                alpha=trgt_dictionary["plotalpha"],
                marker=trgt_dictionary["plotmarker"],
            )


def draw1nomral_segment(points_normal, figure, lenght=100, colors=["g", "y"]):
    # points_normals is a list of 2 elements
    # first element are the normals
    # second element are ponts in space
    starts = points_normal["pivot"]
    normal = points_normal["direction"]  # this ones might not be normalized to 1
    normalized = normal / np.linalg.norm(normal)
    ends = starts + normalized * lenght
    figure.plot(
        [starts[0], ends[0]],
        [starts[1], ends[1]],
        [starts[2], ends[2]],
        color=colors[0],
    )
    figure.scatter(ends[0], ends[1], ends[2], color=colors[1], marker="o")


def draw_nomral_segments(points_normal, figure, lenght=100, colors=["g", "y"], **kwargs):
    # points_normals is a list of 2 elements
    # first element are the normals
    # second element are ponts in space
    starts = points_normal[1]
    normals = points_normal[0]  # this ones might not be normalized to 1

    for i in range(starts.shape[0]):
        normalized = normals[i]
        normalized /= np.linalg.norm(normalized, axis=0)
        ends = starts[i] + normalized * lenght
        figure.plot(
            [starts[i][0], ends[0]],
            [starts[i][1], ends[1]],
            [starts[i][2], ends[2]],
            color=colors[0],
        )
        # figure.scatter(ends[0], ends[1], ends[2], color = 'r', marker = "")
    figure.scatter(
        starts[:, 0], starts[:, 1], starts[:, 2], color=colors[1], marker="o"
    )
    return figure


def stack_projection(stack, angle=45, axes=(1,2), method = "sd", zdepth = False):
    projection = None
    zstack = copy.copy(stack)
    rotated = rotate(zstack, angle=angle, axes=axes, reshape=False, order=3)  # order=3: cubic interpolation
    if zdepth:
        nslices = rotated.shape[2]
        for i in range(1, nslices):
            #intensity_factor = np.log(1.8 + ((nslices - i)/(nslices)))
            intensity_factor = 0.6 * np.exp(((nslices - i)/(nslices)) * (1/(i+1)))
            rotated[:,:,i] = rotated[:,:,i]*intensity_factor
    if method == "max":
        projection = np.max(rotated, axis=2)
    elif method == "sd":
        projection = np.std(rotated, axis=2)
    elif method == "mean":
        projection = np.mean(rotated, axis=2)
    elif method == "sum":
        projection = np.sum(rotated, axis=2)
    return projection

def plot_projection(stack, angle=45, plane="XY", method = "max", zdepth=False):
    if plane == "XY":
        axes = (1, 2)
    elif plane == "XZ": 
        axes = (0, 2)
    elif plane == "YZ":
        axes = (0, 1)
    else:
        raise ValueError("Invalid plane. Choose from 'XY', 'XZ', or 'YZ'.")
    projection = stack_projection(stack, angle=angle, axes=axes, method=method, zdepth=zdepth)
    plt.imshow(projection, cmap='gray')
    plt.axis('off')
    plt.show()


from vlab4mic import experiments
import pytest
import numpy as np
import math


def test_gen_virtual_sample():
    vsample, testexperiment = experiments.generate_virtual_sample()
    assert len(vsample["reference_point"]) == 3


def test_image_output_shape():
    images,images_noiseless, experiment = experiments.image_vsample()
    for modality in images.keys():
        assert type(images[modality]) is dict
        assert len(images[modality]["ch0"].shape) == 3


def test_build_virtual_microscope():
    vmicroscope, experiment = experiments.build_virtual_microscope()


def test_build_vmicroscope_multimodal():
    modalities = ["Widefield", "Confocal", "SMLM", "STED"]
    images, experiment = experiments.build_virtual_microscope(
        multimodal=modalities,
    )
    for modality in modalities:
        assert modality in experiment.imager.modalities.keys()


# structure_list = ["2RCJ", "7R5K", "3J3Y", "1XI5", "1HZH"]
structure_list = [
    "2RCJ",
    "1XI5",
]


@pytest.mark.parametrize("structure", structure_list)
def test_image_sample_structures(structure):
    images,images_noiseless, experiment = experiments.image_vsample(
        structure=structure,
        labelling_efficiency=0.01,
    )
    modalityname = list(images.keys())[0]
    assert len(images[modalityname]["ch0"].shape) == 3


def test_multimodal_imaging():
    modalities = ["Widefield", "Confocal", "SMLM", "STED"]
    images,images_noiseless, experiment = experiments.image_vsample(multimodal=modalities)
    modalityname = list(images.keys())[0]
    assert len(images[modalityname]["ch0"].shape) == 3
    assert len(list(images.keys())) == len(modalities)


def test_download_structure():
    structure9I0K = "9I0K"
    modalities = ["Widefield", "Confocal", "SMLM", "STED"]
    images,images_noiseless, experiment = experiments.image_vsample(
        structure=structure9I0K, multimodal=modalities
    )
    modalityname = list(images.keys())[0]
    assert len(images[modalityname]["ch0"].shape) == 3
    assert len(list(images.keys())) == len(modalities)


def test_vsample_function():
    axial_offset = [0,100]
    yz_orientations = [20,40,60,]
    sample, experiment = experiments.generate_virtual_sample(
        random_orientations=True,
        random_placing=True,
        random_rotations=True,
        rotation_angles=[0,20,30],
        yz_orientations=yz_orientations,
        axial_offset=axial_offset,
        probe_DoL=5,
        number_of_particles=10,
        clear_experiment=True
    )
    z_pos = experiment.coordinate_field.molecules[0].params["ref_point"][2]
    axis = experiment.coordinate_field.molecules[0].axis["direction"]
    original_axis = experiment.coordinate_field.molecules[0].axis_reset["direction"]
    assert z_pos in axial_offset
    original_norm = original_axis / np.linalg.norm(original_axis)
    new_norm = axis / np.linalg.norm(axis)
    dot_product = np.dot(original_norm, new_norm)
    radians=np.arccos(np.clip(dot_product, -1.0, 1.0))
    degree = math.degrees(radians)
    found = False
    for i in yz_orientations:
        if (np.absolute(i - degree)) < 0.00001:
            found = True
    assert found
    experiment.coordinate_field.molecules[0].reset_axis_orientation()
    axis2 = experiment.coordinate_field.molecules[0].axis["direction"]
    axis_diff = axis2 - original_axis
    assert np.linalg.norm(axis_diff) < 0.00001


from .experiments import ExperimentParametrisation
from .analysis import sweep
import matplotlib.pyplot as plt
from IPython.utils import io
import os
from pathlib import Path
from .utils.io.yaml_functions import load_yaml, save_yaml
from .analysis import _plots
import numpy as np
from datetime import datetime
import seaborn as sns
from matplotlib.ticker import FormatStrFormatter
import copy
import tifffile as tiff
from pandas.api.types import is_numeric_dtype
from .analysis.metrics import structural_similarity, pearson_correlation

#output_dir = Path.home() / "vlab4mic_outputs"


class sweep_generator:
    # sweep parameters
    experiment = None
    structures = [
        "1XI5",
    ]
    probes = [
        "NHS_ester",
    ]
    modalities = ["Widefield", "Confocal", "STED", "SMLM"]
    sweep_repetitions = 3
    probe_parameters = None
    structural_integrity_parameters = None
    vsample_parameters = None
    modality_parameters = None
    acquisition_parameters = None

    # reference parameters
    reference_structure = "1XI5"
    reference_probe = "NHS_ester"
    reference_probe_parameters = {"labelling_efficiency": 1.0}
    ####  outputs
    output_directory = None
    reference_virtual_sample = None
    reference_virtual_sample_params = None
    reference_image = None
    reference_image_mask = None
    virtual_samples = None
    virtual_samples_parameters = None
    acquisition_outputs = None
    acquisition_outputs_masks = None
    acquisition_outputs_parameters = None
    # analysis
    analysis = {}
    # analysis["measurements"] = None
    # analysis["inputs"] = None
    # analysis["extended_dataframe"] = None
    analysis["unsorted"] = {}
    analysis["dataframes"] = None
    analysis["plots"] = {}
    # saving
    #output_directory: str = None

    def __init__(self, random_seed: int = None):
        # parameter dictionary
        if random_seed is not None:
            self.experiment = ExperimentParametrisation(random_seed=random_seed)
        else:
            self.experiment = ExperimentParametrisation()
        self.params_by_group = {}
        self.params_by_group["probe"] = {}
        self.params_by_group["virtual_sample"] = {}
        self.params_by_group["particle_structural_integrity"] = {}
        self.params_by_group["modality"] = {}
        self.params_by_group["acquisition"] = {}
        self.output_directory = self.experiment.output_directory
        self.configuration_directory = self.experiment.configuration_path
        param_settings_file = os.path.join(
            self.configuration_directory, "parameter_settings.yaml"
        )
        self.parameter_settings = load_yaml(param_settings_file)
        self.analysis_parameters = {}
        self.analysis_parameters["zoom_in"] = 0
        self.plot_parameters = {}
        self.plot_parameters["ssim"] = {}
        self.plot_parameters["ssim"]["heatmaps"] = {}
        self.plot_parameters["ssim"]["lineplots"] = {}
        self.plot_parameters["pearson"] = {}
        self.plot_parameters["pearson"]["heatmaps"] = {}
        self.plot_parameters["pearson"]["lineplots"] = {}
        self.plot_parameters["heatmaps"] = {}
        self.plot_parameters["heatmaps"]["category"] = "modality_name"
        self.plot_parameters["heatmaps"]["param1"] = None
        self.plot_parameters["heatmaps"]["param2"] = None
        self.plot_parameters["heatmaps"]["annotations"] = False
        self.plot_parameters["heatmaps"]["linewidth"] = 2
        self.plot_parameters["heatmaps"]["palette"] = None
        self.parameters_with_set_values = []
        self.plot_parameters["lineplots"] = {}
        self.plot_parameters["lineplots"]["x_param"] = None
        self.plot_parameters["lineplots"]["hue"] = "modality_name"
        self.plot_parameters["lineplots"]["style"] = None
        self.plot_parameters["lineplots"]["estimator"] = "mean"
        self.plot_parameters["lineplots"]["errorbar"] = "ci"
        self.plot_parameters["general"] = {}
        self.plot_parameters["general"]["na_as_zero"] = True
        self.structures_info_list = self.experiment.structures_info_list
        # Use the directly loaded parameter_settings instead of experiment.param_settings
        # to ensure all parameter groups (including particle_structural_integrity) are available
        self.param_settings = self.parameter_settings
        self.use_experiment_structure = False
        self.reference_parameters_unsorted = dict()
        self.default_metrics = {
            "ssim": structural_similarity,
            "pearson": pearson_correlation
        }
        self.metrics = {}
        
        print("vLab4mic sweep generator initialised")

    def set_number_of_repetitions(self, repeats: int = 3):
        """
        Sets the number of repetitions for each parameter configuration in a sweep.

        Parameters
        ----------
        :param repeats: int, optional
            Number of times to repeat each parameter combination. Defaults to 3.

        Returns
        -------
        None
        """
        self.sweep_repetitions = repeats

    def select_structures(self, structures: list = None, **kwargs):
        """
        Select structures to use in the sweep.

        Parameters
        ----------
        :param structures: list of str, optional
            List of 4-letter PDB/CIF IDs or paths to structure files. If None, uses all available structures.
        :param **kwargs: Additional keyword arguments.

        Returns
        -------
        None
        """
        if structures is not None and type(structures) == list:
            self.structures = structures
            self.use_experiment_structure = False

    def select_probe_templates(self, probe_templates: list = None, **kwargs):
        """
        Select probe templates to use in the sweep.

        Parameters
        ----------
        :param probe_templates: list of str, optional
            List of probe configuration filenames. If None, uses all available probes.
        :param **kwargs: Additional keyword arguments.

        Returns
        -------
        None
        """
        if probe_templates is not None and type(probe_templates) == list:
            self.probes = probe_templates

    def select_modalities(self, modalities: list = None, **kwargs):
        """
        Select imaging modalities to use in the sweep.

        Parameters
        ----------
        :param modalities: list of str, optional
            List of modality names. If None, uses all available modalities.
        :param **kwargs: Additional keyword arguments.

        Returns
        -------
        None
        """
        if modalities is not None and type(modalities) == list:
            self.modalities = modalities

    def set_output_directory(self, output_directory: str = None):
        """
        Set the output directory for saving results.

        Parameters
        ----------
        :param output_directory: str, optional
            Path to the desired output directory. If None, uses the default output directory.

        Returns
        -------
        None
        """
        if output_directory is not None:
            self.output_directory = output_directory

    # generators
    def generate_virtual_samples(self):
        """
        Generate virtual samples from the specified parameter combinations.

        Returns
        -------
        None
        """
        self.create_parameters_iterables()
        if self.use_experiment_structure:
                       (
                self.experiment,
                self.virtual_samples,
                self.virtual_samples_parameters,
            ) = sweep.sweep_vasmples(
                experiment=self.experiment,
                structures=self.structures,
                probes=self.probes,
                probe_parameters=self.probe_parameters,
                particle_structural_integrity=self.structural_integrity_parameters,
                virtual_samples=self.vsample_parameters,
                repetitions=self.sweep_repetitions,
                use_experiment_structure=True
            )
        else: 
            (
                self.experiment,
                self.virtual_samples,
                self.virtual_samples_parameters,
            ) = sweep.sweep_vasmples(
                experiment=self.experiment,
                structures=self.structures,
                probes=self.probes,
                probe_parameters=self.probe_parameters,
                particle_structural_integrity=self.structural_integrity_parameters,
                virtual_samples=self.vsample_parameters,
                repetitions=self.sweep_repetitions,
            )

    def generate_acquisitions(self):
        """
        Generate image simulation acquisition for all virtual samples generated with generate_virtual_samples.

        Returns
        -------
        None
        """
        # generate virtual samples
        with io.capture_output() as captured:
            if self.virtual_samples is None:
                self.generate_virtual_samples()
        # acquisition of virtual samples
        (
            self.experiment,
            self.acquisition_outputs,
            self.acquisition_outputs_parameters,
            mod_pixelsizes,
            self.acquisition_outputs_masks
        ) = sweep.sweep_modalities_updatemod(
            experiment=self.experiment,
            vsample_outputs=self.virtual_samples,
            vsampl_pars=self.virtual_samples_parameters,
            modalities=self.modalities,
            modality_params=self.modality_parameters,
            modality_acq_prams=self.acquisition_parameters,
        )

    def set_reference_parameters(
        self,
        reference_structure: str = None,
        reference_probe: str = None,
        reference_probe_parameters: dict = None,
        **kwargs,
    ):
        """
        Specify parameters to use for both reference sample and reference image used at analysis.

        Parameters
        ----------
        :param reference_structure: str, optional
            4-letter ID of PDB/CIF model.
        :param reference_probe: str, optional
            Name ID of probe configuration file (filename).
        :param reference_probe_parameters: dict, optional
            Probe parameters to use.
        :param **kwargs: Additional keyword arguments.

        Returns
        -------
        None
        """
        if reference_structure is None:
            self.reference_structure = self.structures[0]
        else:
            self.reference_structure = reference_structure
        if reference_probe is not None:
            self.reference_probe = reference_probe
        if reference_probe_parameters is not None:
            self.reference_probe_parameters = reference_probe_parameters
        if "relative_positions" in kwargs.keys():
            self.reference_parameters_unsorted["particle_positions"] = kwargs.pop("relative_positions")
        for key, value in kwargs.items():
            self.reference_parameters_unsorted[key] = value

    def generate_reference_sample(self, use_experiment_probe=True, use_experiment_vsample=True):
        """
        Generate the reference virtual sample and its parameters.

        Returns
        -------
        None
        """
        if self.use_experiment_structure:
            structure_is_path = True
        else: 
            structure_is_path = False
        if use_experiment_probe:
            reference_probe = list(self.probes)[0]
        else:
            reference_probe = None
        if use_experiment_vsample:
            reference_sample_params = copy.copy(self.experiment.virtualsample_params)
        self.set_reference_parameters(
            reference_structure=self.reference_structure,
            reference_probe=reference_probe,
            **reference_sample_params
            )
        self.reference_virtual_sample, self.reference_virtual_sample_params = (
            sweep.generate_global_reference_sample(
                structure=self.reference_structure,
                probe=self.reference_probe,
                probe_parameters=self.reference_probe_parameters,
                structure_is_path=structure_is_path,
                **self.reference_parameters_unsorted
            )
        )

    def generate_reference_image(self, override=False):
        """
        Generate the reference image using the reference virtual sample.

        Parameters
        ----------
        :param override: bool, optional
            If True, regenerate the reference image even if it exists. Defaults to False.

        Returns
        -------
        None
        """
        if self.reference_image is None or override:
            if self.use_experiment_structure:
                self.reference_structure = copy.copy(self.structure_path)
                self.generate_reference_sample()
            else:
                self.generate_reference_sample()
        self.reference_image, self.reference_image_parameters, self.reference_image_mask = (
            sweep.generate_global_reference_modality(
                reference_vsample=self.reference_virtual_sample,
                reference_vsample_params=self.reference_virtual_sample_params,
            )
        )

    def load_reference_image(
        self, ref_image_path=None, ref_pixelsize=None, ref_image_mask_path = None, override=True
    ):
        """
        Load a reference image from a specified path.
        """
        reference_parameters = {}
        reference_parameters["Vector"] = None
        ref_pixelsize
        reference_parameters["ref_pixelsize"] = ref_pixelsize
        ref_image = tiff.imread(ref_image_path)
        if ref_image_mask_path is not None:
            image_mask = tiff.imread(ref_image_mask_path)
            ref_image_mask = image_mask > 0
        else:
            if len(ref_image.shape) == 3:
                image_mask = np.ones(shape=ref_image[0].shape)
            else:
                image_mask = np.ones(shape=ref_image.shape)
            ref_image_mask = image_mask > 0
        if override:
            self.reference_image = ref_image
            self.reference_image_parameters = reference_parameters
            self.reference_image_mask = ref_image_mask
            

    # previews
    def preview_image_output_by_ID(
        self,
        probe_template=0,
        probe_parameters=0,
        structural_integrity_parameters=0,
        virtual_sample_parameters=0,
        modality_template=0,
        modality_parameters=0,
        acquisition_parameters=0,
        replica_number=0,
        frame=0,
        return_image=False,
        cmap="Grays_r"
    ):
        """
        Preview or return the first acquisition output image.

        Parameters
        ----------
        :param return_image: bool, optional
            If True, return the image array instead of displaying it. Defaults to False.

        Returns
        -------
        numpy.ndarray or None
            The image array if `return_image` is True, otherwise None.
        """
        parameter_id = (
            str(probe_template)
            + "_"
            + str(probe_parameters)
            + "_"
            + str(structural_integrity_parameters)
            + "_"
            + str(virtual_sample_parameters)
            + "_"
            + str(modality_template)
            + "_"
            + str(modality_parameters)
            + "_"
            + str(acquisition_parameters)
        )
        if (
            parameter_id is None
            or parameter_id not in self.acquisition_outputs_parameters.keys()
        ):
            # parameter_id = list(self.acquisition_outputs_parameters.keys())[0]
            return None
        if (
            len(self.acquisition_outputs[parameter_id][replica_number].shape)
            == 3
        ):
            image = self.acquisition_outputs[parameter_id][replica_number][
                frame
            ]
        else:
            image = self.acquisition_outputs[parameter_id][replica_number]
        if return_image:
            return image, self.acquisition_outputs_parameters[parameter_id]
        else:
            plt.imshow(image, cmap=cmap)
            print(self.acquisition_outputs_parameters[parameter_id])

    def preview_image_output_by_parameter_values(self, **kwargs):
        pass

    def preview_reference_image(self, return_image=False, cmap="Grays_r"):
        """
        Preview or return the reference image.

        Parameters
        ----------
        :param return_image: bool, optional
            If True, return the image array instead of displaying it. Defaults to False.

        Returns
        -------
        numpy.ndarray or None
            The image array if `return_image` is True, otherwise None.
        """
        if return_image:
            return self.reference_image
        else:
            if len(self.reference_image.shape) > 2:
                plt.imshow(self.reference_image[0], cmap=cmap)
            else:
                plt.imshow(self.reference_image, cmap=cmap)
            print(self.reference_image_parameters)

    # set and change parameters
    def set_parameter_values(self, param_group, param_name, values=None):
        """
        Specify list or range of values to use per parameter.

        Parameters
        ----------
        :param param_group: str
            Group of parameter, e.g., 'probe', 'virtual_sample'.
        :param param_name: str
            Name of parameter to use for sweep.
        :param values: None, tuple, or list, optional
            Parameter values to use. If tuple, interpreted as (start, stop, num) for linspace.
            If list, used directly. If None, defaults are generated.

        Returns
        -------
        None
        """
        if param_group in list(self.params_by_group.keys()):
            if values is None:
                # create defaults or ignore this parameter for it to not be used
                pass
            elif type(values) == list:
                # set list directly as the params to use
                self.params_by_group[param_group][param_name] = values
            elif type(values) == tuple:
                # 3 values are expected: min, max, steps
                # generate a linspace
                if param_group in self.parameter_settings.keys():
                    if (
                        param_name
                        in self.parameter_settings[param_group].keys()
                    ):
                        if (
                            self.parameter_settings[param_group][param_name][
                                "wtype"
                            ]
                            == "int_slider"
                        ):
                            #step = np.ceil((values[1] - values[0]) / values[2])
                            self.params_by_group[param_group][param_name] = (
                                np.arange(
                                    start=values[0],
                                    stop=values[1],
                                    step=values[2],
                                    dtype=int,
                                )
                            )
                        else:
                            num = int((values[1]-values[0]) / values[2]) + 1
                            param_iterables = np.linspace(
                                values[0], values[1], num
                            )
                            self.params_by_group[param_group][
                                param_name
                            ] = param_iterables
            self.parameters_with_set_values.append(param_name)
        else:
            print(f"{param_group} is not a valid parameter group")

    def set_sweep_parameters(
            self,
            # probe
            probe_target_type = None,
            probe_target_value = None,
            probe_target_option= None,
            probe_model = None,
            probe_fluorophore = None,
            probe_paratope = None,
            probe_conjugation_target_info=None,
            probe_seconday_epitope=None,
            peptide_motif: dict = None,
            probe_distance_to_epitope= None,
            probe_steric_hindrance= None,
            probe_DoL = None,
            probe_wobble_theta = None,
            labelling_efficiency = None,
            # structural_integrity
            structural_integrity=None,
            structural_integrity_small_cluster=None,
            structural_integrity_large_cluster=None,
            # vsample
            sample_dimensions = None,
            particle_positions = None,
            particle_orientations = None,
            rotation_angles = None,
            minimal_distance = None,
            # modality params
            pixelsize_nm = None,
            lateral_resolution_nm = None,
            axial_resolution_nm = None,
            psf_voxel_nm = None,
            depth_of_field_nm = None,
            # imaging 
            exp_time=None,
            **kwargs
    ):
        if probe_distance_to_epitope is not None:
            self.set_parameter_values(
                "probe", "probe_distance_to_epitope", values=probe_distance_to_epitope
            )
        if probe_steric_hindrance is not None:
            self.set_parameter_values(
                "probe", "probe_steric_hindrance", values=probe_steric_hindrance
            )
        if probe_DoL is not None:
            self.set_parameter_values(
                "probe", "probe_DoL", values=probe_DoL
            )
        if probe_wobble_theta is not None:
            self.set_parameter_values(
                "probe", "probe_wobble_theta", values=probe_wobble_theta
            )
        if labelling_efficiency is not None:
            self.set_parameter_values(
                "probe", "labelling_efficiency", values=labelling_efficiency
            )
        if probe_target_type is not None:
            self.set_parameter_values(
                "probe", "probe_target_type", values=probe_target_type
            )
        if probe_target_value is not None:
            self.set_parameter_values(
                "probe", "probe_target_value", values=probe_target_value
            )
        if probe_target_option is not None:
            self.set_parameter_values(
                "probe", "probe_target_option", values=probe_target_option
            )
        if probe_model is not None:
            self.set_parameter_values(
                "probe", "probe_model", values=probe_model
            )
        if probe_fluorophore is not None:
            self.set_parameter_values(
                "probe", "probe_fluorophore", values=probe_fluorophore
            )
        if probe_paratope is not None:
            self.set_parameter_values(
                "probe", "probe_paratope", values=probe_paratope
            )
        if probe_conjugation_target_info is not None:
            self.set_parameter_values(
                "probe", "probe_conjugation_target_info", values=probe_conjugation_target_info
            )
        if probe_seconday_epitope is not None:
            self.set_parameter_values(
                "probe", "probe_secondary_epitope", values=probe_seconday_epitope
            )

        if structural_integrity is not None:
            self.set_parameter_values(
                "particle_structural_integrity", "structural_integrity", values=structural_integrity
            )
        if structural_integrity_small_cluster is not None:
            self.set_parameter_values(
                "particle_structural_integrity", "structural_integrity_small_cluster", values=structural_integrity_small_cluster
            )
        if structural_integrity_large_cluster is not None:
            self.set_parameter_values(
                "particle_structural_integrity", "structural_integrity_large_cluster", values=structural_integrity_large_cluster
            )
        if sample_dimensions is not None:
            self.set_parameter_values(
                "virtual_sample", "sample_dimensions", values=sample_dimensions
            )
        if particle_positions is not None:
            self.set_parameter_values(
                "virtual_sample", "particle_positions", values=particle_positions
            )
        if particle_orientations is not None:
            self.set_parameter_values(
                "virtual_sample", "particle_orientations", values=particle_orientations
            )
        if rotation_angles is not None:
            self.set_parameter_values(
                "virtual_sample", "rotation_angles", values=rotation_angles
            )
        if exp_time is not None:
            self.set_parameter_values(
                "acquisition", "exp_time", values=exp_time
            )
        if pixelsize_nm is not None:
            self.set_parameter_values(
                "modality", "pixelsize_nm", values=pixelsize_nm
            )
        if lateral_resolution_nm is not None:
            self.set_parameter_values(
                "modality", "lateral_resolution_nm", values=lateral_resolution_nm
            )
        if axial_resolution_nm is not None:
            self.set_parameter_values(
                "modality", "axial_resolution_nm", values=axial_resolution_nm
            )
        if psf_voxel_nm is not None:
            self.set_parameter_values(
                "modality", "psf_voxel_nm", values=psf_voxel_nm
            )
        if depth_of_field_nm is not None:
            self.set_parameter_values(
                "modality", "depth_of_field_nm", values=depth_of_field_nm
            )
        if peptide_motif is not None:
            pass
        

    def clear_sweep_parameters(self):
        """
        Clear all parameters set for the sweep, resetting to default values.

        Returns
        -------
        None
        """
        for group_name in self.params_by_group.keys():
            self.params_by_group[group_name] = {}
        self.parameters_with_set_values = []

    def create_parameters_iterables(self):
        """
        Create iterables for all parameter groups based on set values.

        Returns
        -------
        None
        """
        param_groups = list(self.params_by_group.keys())
        no_params_set = True
        for group_name in param_groups:
            if len(self.params_by_group[group_name]) > 0:
                no_params_set = False
        if no_params_set:
            # set default with minimal options to iterate
            self.set_parameter_values(
                "probe", "labelling_efficiency", values=[0.5, 1]
            )

        self.probe_parameters = sweep.create_param_combinations(
            **self.params_by_group["probe"]
        )
        self.structural_integrity_parameters = sweep.create_param_combinations(
            **self.params_by_group["particle_structural_integrity"]
        )
        self.vsample_parameters = sweep.create_param_combinations(
            **self.params_by_group["virtual_sample"]
        )
        self.acquisition_parameters = sweep.create_param_combinations(
            **self.params_by_group["acquisition"]
        )
        self.modality_parameters = sweep.create_param_combinations(
            **self.params_by_group["modality"]
        )

    def set_analysis_parameters(
        self, metrics_list: list[str] = None, zoom_in: float = None, **kwargs
    ):
        """
        Set analysis parameters for the simulation.

        Parameters
        ----------
        :param metrics_list: list of str, optional
            A list of metric names to be used in the analysis.
        :param zoom_in: float, optional
            A zoom factor to be applied during analysis.
        :param **kwargs: Additional keyword arguments for future extensibility.

        Returns
        -------
        None
        """
        #if metrics_list is not None and type(metrics_list) == list:
        #    self.analysis_parameters["metrics_list"] = metrics_list
        if zoom_in is not None:
            self.analysis_parameters["zoom_in"] = zoom_in

    def set_plot_parameters(self, plot_type, **kwargs):
        """
        Set plotting parameters for a given plot type.

        Parameters
        ----------
        :param plot_type: str
            The type of plot to configure.
        :param **kwargs: Plot-specific keyword arguments.

        Returns
        -------
        None
        """
        if plot_type in self.plot_parameters.keys():
            for key, val in kwargs.items():
                self.plot_parameters[plot_type][key] = val

    def set_na_as_zero_in_plots(self, na_as_zero: bool = True):
        """
        Set whether to treat NaN values as zero in plots.

        Parameters
        ----------
        :param na_as_zero: bool, optional
            If True, NaN values will be treated as zero in plots. Defaults to True.
            This does not affect the underlying data, only the visualization.

        Returns
        -------
        None
        """
        self.plot_parameters["general"]["na_as_zero"] = na_as_zero   

    def use_default_metrics(self, metrics = ["ssim", "pearson"]):
        if metrics is not None:
            for metric_name in metrics:
                self.metrics[metric_name] = self.default_metrics[metric_name]

    def run_analysis(
        self,
        save=True,
        output_name=None,
        output_directory=None,
        plots=False,
        save_images=True,
        **kwargs,
    ):
        """
        Analyse image simulations against the specified image reference.

        This method generates a dataframe containing the calculated metrics per parameter combination.

        Parameters
        ----------
        :param save: bool, optional
            Specify if output analysis will be saved in output_directory. Defaults to True.
        :param output_name: str, optional
            Name to identify analysis output files.
        :param output_directory: str, optional
            Path for writing outputs.
        :param plots: bool, optional
            Generate heatmaps and lineplots from the results dataframe. Defaults to False.
        :param save_images: bool, optional
            Whether to save images. Defaults to True.
        :param **kwargs: Additional keyword arguments.

        Returns
        -------
        None
        """
        if self.acquisition_outputs is None:
            self.generate_acquisitions()
        if self.reference_image is None:
            self.generate_reference_image()
        if len(self.reference_image.shape) == 3:
            # if reference image is 3D, take the first frame
            reference_image = self.reference_image[0]
            reference_image_mask = self.reference_image_mask
        else:
            reference_image = self.reference_image
            reference_image_mask = self.reference_image_mask
        if len(self.metrics.keys()) == 0:
            self.use_default_metrics() 
        print("Running analysis...")
        measurement_vectors, inputs, metric = (
            sweep.analyse_sweep_single_reference(
                img_outputs=self.acquisition_outputs,
                img_outputs_masks=self.acquisition_outputs_masks,
                img_params=self.acquisition_outputs_parameters,
                reference_image=reference_image,
                reference_image_mask=reference_image_mask,
                reference_params=self.reference_image_parameters,
                metrics=self.metrics,
                **self.analysis_parameters,
            )
        )
        print("Analysis complete.")
        self.analysis["unsorted"]["measurement_vectors"] = measurement_vectors
        self.analysis["unsorted"]["inputs"] = inputs
        print("Generating analysis dataframe...")
        self.gen_analysis_dataframe()
        print("Analysis dataframe generated.")
        if plots:
            print("Generating analysis plots...")
            for metric_name in self.metrics.keys():
                for plot_type in ["heatmaps", "lineplots"]:
                    self.generate_analysis_plots(
                        plot_type=plot_type,
                        return_figure=True,
                        metric_name=metric_name,
                        filter_dictionary=None,
                        na_as_zero=self.plot_parameters["general"]["na_as_zero"],
                        **self.plot_parameters[metric_name][plot_type]
                    )
            print("Analysis plots generated.")
        if save:
            print("saving analysis...")
            self.save_analysis(
                output_name=output_name, output_directory=output_directory
            )
            print("Analysis saved.")
        print("Analysis run complete.")

    def gen_analysis_dataframe(self):
        """
        Generate the analysis dataframe from measurement vectors and parameters.

        Returns
        -------
        None
        """
        unformatted_df, self.analysis["dataframes"] = (
            sweep.measurements_dataframe(
                measurement_vectors=self.analysis["unsorted"][
                    "measurement_vectors"
                ],
                probe_parameters=self.probe_parameters,
                p_structural_integrity=self.structural_integrity_parameters,
                sample_params=self.vsample_parameters,
                mod_acq=self.acquisition_parameters,
                mod_names=self.modalities,
                mod_params=self.modality_parameters,
                metric_names=list(self.metrics.keys()),
            )
        )

    # methods to retrieve attributes
    def get_analysis_output(self, keyname="dataframes"):
        """
        Retrieve the analysis output associated with the specified key.

        Parameters
        ----------
        :param keyname: str, optional
            The key corresponding to the desired analysis output. Defaults to "dataframes".

        Returns
        -------
        Any
            The analysis output associated with the given key.

        Raises
        ------
        KeyError
            If the specified key does not exist in the analysis dictionary.
        """
        return self.analysis[keyname]

    def generate_analysis_plots(
        self,
        plot_type=None,
        metric_name=None,
        decimals: int = None,
        return_figure=True,
        filter_dictionary=None,
        na_as_zero = True,
        **kwargs,
    ):
        """
        Generates and stores analysis plots based on the specified plot type and metric.

        Parameters
        ----------
        :param plot_type: str, optional
            The type of plot to generate. Supported types are "heatmaps" and "lineplots".
            If not specified, defaults to "heatmaps".
        :param metric_name: str, optional
            The name of the metric to plot. Required for generating the plot.
        :param decimals: int, optional
            Number of decimal places to format metric values. If not specified, defaults to 3.
        :param **kwargs: Additional keyword arguments passed to the underlying plotting functions.

        Notes
        -----
        The generated plot is stored in the `self.analysis["plots"]` dictionary under the corresponding plot type and metric name.
        Plot parameters are retrieved from `self.plot_parameters` based on the plot type.
        Requires that analysis output dataframes are available via `self.get_analysis_output(keyname="dataframes")`.

        Returns
        -------
        None
        """
        if decimals is None:
            decimals = "%.3f"
        else:
            decimals = "%." + str(decimals) + "f"
        data = self.get_analysis_output(keyname="dataframes")
        if plot_type is None:
            plot_type = "heatmaps"
        if plot_type not in self.analysis["plots"].keys():
            self.analysis["plots"][plot_type] = {}
        plot_params = self.plot_parameters[plot_type]
        if plot_type == "heatmaps":
            metric_plot = self._gen_heatmaps(
                metric_name=metric_name,
                return_figure=return_figure,
                decimals=decimals,
                filter_dictionary=filter_dictionary,
                na_as_zero = na_as_zero,
                **plot_params,
                **kwargs
            )
            self.analysis["plots"][plot_type][metric_name] = metric_plot
        elif plot_type == "lineplots":
            metric_plot = self._gen_lineplots(
                data=data,
                metric_name=metric_name,
                decimals=decimals,
                return_figure=return_figure,
                filter_dictionary=filter_dictionary,
                na_as_zero = na_as_zero,
                **plot_params,
                **kwargs
            )
            self.analysis["plots"][plot_type][metric_name] = metric_plot

    def _gen_heatmaps(
        self,
        metric_name=None,
        category: str = None,
        param1: str = None,
        param2: str = None,
        return_figure=False,
        decimals="%.4f",
        filter_dictionary=None,
        annotations=False,
        palette=None,
        na_as_zero = False,
        **kwargs,
    ):
        """
        Generate heatmap plots for the specified metric and parameters.

        Parameters
        ----------
        :param metric_name: str, optional
            Name of the metric to plot.
        :param category: str, optional
            Category for pivoting data (e.g., modality_name).
        :param param1: str, optional
            First parameter for the heatmap axes.
        :param param2: str, optional
            Second parameter for the heatmap axes.
        :param return_figure: bool, optional
            If True, return the matplotlib figure. Defaults to False.
        :param decimals: str, optional
            Format string for decimal places. Defaults to '%.4f'.
        :param **kwargs: Additional keyword arguments.

        Returns
        -------
        matplotlib.figure.Figure
            The generated heatmap figure.
        """
        if metric_name is None:
            metric_name = list(self.metrics.keys())[0]
        if category is None:
            category = "modality_name"
        if param1 is None:
            param1 = self.parameters_with_set_values[0]
        if param2 is None:
            param_names_set = self.parameters_with_set_values
            if len(param_names_set) >= 2:
                param2=self.parameters_with_set_values[1]
            else:
                param2 = "probe_n"
        analysis_resut_df = self.get_analysis_output(keyname="dataframes")
        df = copy.deepcopy(self.analysis["dataframes"])
        if is_numeric_dtype(df[param1]):
            df[param1] = df[param1].round(3)
        if is_numeric_dtype(df[param2]):
            df[param2] = df[param2].round(3)
        pre_filter_dt = True
        if filter_dictionary is not None:
             pre_filter_dt = True
        df_categories, titles = sweep.pivot_dataframes_byCategory(
            dataframe=df,
            category_name=category,
            param1=param1,
            param2=param2,
            metric_name=metric_name,
            pre_filter_dt=pre_filter_dt,
            filter_dictionary=filter_dictionary,
            **kwargs
        )
        conditions = list(df_categories.keys())
        nconditions = len(conditions)
        if palette is None:
            cmap_palette = sns.diverging_palette(20, 200, s=80, l=50, as_cmap=True)
        else: 
            cmap_palette = palette
        plot = _plots.sns_heatmap_pivots(
            df_categories,
            titles,
            annotations=annotations,
            return_figure=return_figure,
            metric_name=metric_name,
            conditions_cmaps=[cmap_palette]*nconditions,
            decimals=decimals,
            na_as_zero = na_as_zero,
            **kwargs
        )
        return plot

    def _gen_lineplots(
        self,
        data=None,
        x_param=None,
        metric_name=None,
        hue="modality_name",
        style=None,
        markers=None,
        estimator="mean",
        errorbar="ci",
        figsize=[10, 10],
        decimals="%.4f",
        return_figure=True,
        filter_dictionary=None,
        na_as_zero = False,
        **kwargs,
    ):
        """
        Generate line plots for the specified metric and parameters.

        Parameters
        ----------
        :param data: pandas.DataFrame, optional
            Data to plot. If None, uses self.analysis["dataframes"].
        :param x_param: str, optional
            Parameter for the x-axis.
        :param metric_name: str, optional
            Metric to plot on the y-axis.
        :param hue: str, optional
            Variable that defines subsets of the data, which will be drawn on separate lines.
        :param style: str, optional
            Variable that determines the line style.
        :param markers: bool or list, optional
            Markers for the lines.
        :param estimator: str or callable, optional
            Method for aggregating data.
        :param errorbar: str, optional
            Error bar type.
        :param figsize: list, optional
            Figure size.
        :param decimals: str, optional
            Format string for decimal places.
        :param **kwargs: Additional keyword arguments.

        Returns
        -------
        matplotlib.figure.Figure
            The generated line plot figure.
        """
        if data is None:
            data = self.analysis["dataframes"]
        if x_param is None and len(self.parameters_with_set_values) > 0:
            x_param = self.parameters_with_set_values[0]
        else:
            x_param = "labelling_efficiency"
        if metric_name is None:
            metric_name = list(self.metrics.keys())[0]
        if style is None and len(self.parameters_with_set_values) > 1:
            style = self.parameters_with_set_values[1]
        fig, axes = plt.subplots(figsize=figsize)
        if na_as_zero:
            data_to_plot = data.copy(deep=True)
            data_to_plot.fillna(0, inplace=True)
        else:
            data_to_plot = data
        sns.lineplot(
            data=data_to_plot,
            x=x_param,
            y=metric_name,
            hue=hue,
            style=style,
            markers=markers,
            estimator=estimator,
            errorbar=errorbar,
            ax=axes,
            **kwargs
        )
        axes.yaxis.set_major_formatter(FormatStrFormatter(decimals))
        axes.xaxis.set_major_formatter(FormatStrFormatter(decimals))
        title = estimator + " " + metric_name + " for " + x_param
        if style is not None:
            title = title + " per " + style
        plt.title(title)
        plt.close()
        return fig

    def save_analysis(
        self, output_name=None, output_directory=None, analysis_type=None
    ):
        """
        Saves analysis results, including dataframes and plots, to the specified output directory.

        Parameters
        ----------
        :param output_name: str, optional
            Base name for the output files. If None, defaults to "vLab4mic_results_".
            The current date (YYYYMMDD) will be appended to the name.
        :param output_directory: str, optional
            Directory where the output files will be saved. If None, uses the instance's `output_directory` attribute.
        :param analysis_type: list of str, optional
            Types of analysis outputs to save. Can include "dataframes" and/or "plots".
            If None, both "dataframes" and "plots" are saved.

        Notes
        -----
        Dataframes are saved as CSV files.
        Plots are saved as PNG files, named according to their metric and plot type.
        The method relies on `get_analysis_output` to retrieve the relevant dataframes and plots.

        Returns
        -------
        None
        """
        now = datetime.now()  # dd/mm/YY H:M:S
        dt_string = now.strftime("%Y%m%d")
        if analysis_type is None:
            analysis_type = ["dataframes", "plots"]
        if output_name is None:
            output_name = "vLab4mic_results_"
        output_name = output_name + dt_string
        if output_directory is None:
            output_directory = self.output_directory
        for keyname in analysis_type:
            if keyname == "dataframes":
                df = self.get_analysis_output(keyname)
                df_name = output_name + "_dataframe.csv"
                files_indir = os.listdir(output_directory)
                if df_name in files_indir:
                    count = 1
                    df_name = (
                        output_name + "_dataframe_" + str(count) + ".csv"
                    )
                    while df_name in files_indir:
                        count += 1
                        df_name = (
                            output_name + "_dataframe_" + str(count) + ".csv"
                        )
                df.to_csv(os.path.join(output_directory, df_name), index=False)
            elif keyname == "plots":
                plots_dictionary = self.get_analysis_output(keyname)
                for plot_type, plots_by_metric in plots_dictionary.items():
                    for metric, plot in plots_by_metric.items():
                        figure_name = (
                            output_name
                            + "_"
                            + metric
                            + "_"
                            + plot_type
                            + ".png"
                        )
                        current_files = os.listdir(output_directory)
                        if figure_name in current_files:
                            count = 1
                            figure_name = (
                                output_name
                                + "_"
                                + metric
                                + "_"
                                + plot_type
                                + "_"
                                + str(count)
                                + ".png"
                            )
                            while figure_name in current_files:
                                count += 1
                                figure_name = (
                                    output_name
                                    + "_"
                                    + metric
                                    + "_"
                                    + plot_type
                                    + "_"
                                    + str(count)
                                    + ".png"
                                )
                        plot.savefig(
                            os.path.join(output_directory, figure_name)
                        )

    def save_images(self, output_name=None, output_directory=None, floats_as=float):
        """
        Save simulated acquisition images and the reference image to disk in TIFF format.

        Parameters
        ----------
        :param output_name: str, optional
            Base name for the output image files. Defaults to "vLab4mic_images_".
        :param output_directory: str, optional
            Directory where images will be saved. If not provided, uses a default path
            based on `self.output_directory` under a "simulated_images" subdirectory. The directory is created if it does not exist.
        :param floats_as: callable, optional
            Function to convert numpy float64 values in parameter dictionaries to another type (e.g., float). Defaults to float.

        Returns
        -------
        None

        Notes
        -----
        - Iterates over all parameter combinations in `self.acquisition_outputs`.
        - For each combination, concatenates all replicate images and saves the result as a TIFF file named after the parameter combination ID.
        - If `self.reference_image` is present, saves it as "reference.tiff" in the output directory.
        - Saves acquisition parameters as a YAML file in the output directory.
        """
        now = datetime.now()  # dd/mm/YY H:M:S
        dt_string = now.strftime("%Y%m%d")
        if output_name is None:
            output_name = "vLab4mic_images_"
        if output_directory is None:
            foldername = "simulated_images_" + dt_string
            filesindir = os.listdir(self.output_directory)
            if foldername in filesindir:
                count = 1
                foldername = "simulated_images_" + dt_string + "_" + str(count)
                while foldername in filesindir:
                    count+=1
                    foldername = "simulated_images_" + dt_string + "_" + str(count)
            output_directory = os.path.join(
                self.output_directory, foldername
            )
            if not os.path.exists(output_directory):
                os.makedirs(output_directory)
        for (
            param_combination_id,
            replicates,
        ) in self.acquisition_outputs.items():
            nreps = len(replicates)
            image = replicates[0]
            for i in range(1, nreps):
                image = np.concatenate((image, replicates[i]))
            name = param_combination_id + ".tiff"
            dir_name = os.path.join(output_directory, name)
            tiff.imwrite(dir_name, image)
        if floats_as is not None and callable(floats_as):
            copy_of_params = copy.deepcopy(self.acquisition_outputs_parameters)
            for combination_id, list_of_parameters in copy_of_params.items():
                for parameter in list_of_parameters:
                    if type(parameter) is dict:
                        for parameter_name in parameter.keys():
                            if type(parameter[parameter_name]) is np.float64:
                                parameter[parameter_name] = floats_as(parameter[parameter_name])
            save_yaml(
                data=copy_of_params,
                name="acquisition_parameters",
                output_directory=output_directory,
            )            
        else:
            save_yaml(
                data=self.acquisition_outputs_parameters,
                name="acquisition_parameters",
                output_directory=output_directory,
            )
        if self.reference_image is not None:
            # save reference image
            name_ref = param_combination_id + ".tiff"
            dir_name_ref = os.path.join(output_directory, name_ref)
            #name_ref = output_directory + "reference.tiff"
            tiff.imwrite(dir_name_ref, self.reference_image)

    def add_custom_analysis_metrics(self, custom_metrics: list[callable] = None, heatmap_params={"cmaps_range": "each"}, lineplots_params=None, **kwargs):
        """
        Add a custom analysis metric function to the sweep generator.

        Parameters
        ----------
        :param metric_function: callable
            A function that takes two numpy arrays (reference image and query image)
            and returns a float representing the calculated metric.
        :param metric_name: str
            The name of the custom metric to be added.

        Returns
        -------
        None

        Notes
        -----
        The custom metric function should have the following signature:
            def custom_metric(ref: np.ndarray, query: np.ndarray, union_mask, **kwargs) -> float:
                # here union_mask is a binary mask that can be used to filter pixel to use for calculations
                # Calculate and return the metric value
        """
        current_metrics = list(self.metrics.keys())
        for m in range(len(custom_metrics)):
            metric_name = custom_metrics[m].__name__
            if metric_name not in current_metrics:
                #self.analysis_parameters["metrics_list"].append(metric_name)
                self.plot_parameters[metric_name] = dict()
            # change plot parameters to different method
            if heatmap_params is not None:
                self.plot_parameters[metric_name]["heatmaps"] = heatmap_params
            else:
                self.plot_parameters[metric_name]["heatmaps"] = {}
            if lineplots_params is not None:
                self.plot_parameters[metric_name]["lineplots"] = lineplots_params
            else:
                self.plot_parameters[metric_name]["lineplots"] = {}
            self.metrics[metric_name] = custom_metrics[m]
        


def run_parameter_sweep(
    structures: list[str] = None,
    probe_templates: list[str] = None,
    modalities: list[str] = None,
    output_directory: str = None,
    output_name: str = None,
    save_analysis_results: bool = True,
    analysis_plots: bool = True,
    save_sweep_images: bool = True,
    sweep_repetitions: int = 3,
    return_generator: bool = False,
    number_of_particles: int = 1,
    particle_positions: int = None,
    reference_structure = None,
    reference_probe = None,
    reference_parameters: dict = None,
    clear_experiment=True,
    run_analysis=True,
    # parameters to sweep
    probe_target_type = None,
    probe_target_value = None,
    probe_target_option = None,
    probe_model = None,
    probe_fluorophore = None,
    probe_paratope = None,
    probe_conjugation_target_info = None,
    probe_seconday_epitope = None,
    peptide_motif: dict = None,
    probe_distance_to_epitope = None,
    probe_steric_hindrance = None,
    probe_DoL = None,
    probe_wobble_theta = None,
    labelling_efficiency = None,
    structural_integrity = None,
    structural_integrity_small_cluster = None,
    structural_integrity_large_cluster = None,
    sample_dimensions = None,
    particle_orientations = None,
    rotation_angles = None,
    minimal_distance = None,
    pixelsize_nm = None,
    lateral_resolution_nm = None,
    axial_resolution_nm = None,
    psf_voxel_nm = None,
    depth_of_field_nm = None,
    exp_time = None,
    # for plot generation
    na_as_zero = True,
    custom_metrics: list[callable] = None,
    default_metrics =  ["ssim", "pearson"],
    #custom_metric_name: str = None,
    plot_parameters=None,
    random_seed: int = None
    # Add more as needed for your sweep
):
    """
    Run a parameter sweep for virtual microscopy simulations and analysis.

    This function automates the process of generating virtual samples, simulating
    acquisitions, and performing analysis across a range of user-specified parameters.
    It supports batch processing of multiple structures, probes, modalities, and
    imaging/acquisition settings, and can optionally save results and images.

    Parameters
    ----------
    :param structures: list of str, optional
        List of structure IDs or file paths to use in the sweep.
    :param probe_templates: list of str, optional
        List of probe configuration filenames to use in the sweep.
    :param modalities: list of str, optional
        List of imaging modality names to use in the sweep.
    :param output_directory: str, optional
        Directory to save outputs. If None, uses default.
    :param output_name: str, optional
        Base name for output files.
    :param save_analysis_results: bool, optional
        Whether to save analysis results. Default is True.
    :param analysis_plots: bool, optional
        Whether to generate and save analysis plots. Default is True.
    :param save_sweep_images: bool, optional
        Whether to save simulated images. Default is True.
    :param sweep_repetitions: int, optional
        Number of repetitions per parameter combination. Default is 3.
    :param return_generator: bool, optional
        If True, return the sweep_generator instance. Default is False.
    :param number_of_particles: int, optional
        Number of particles per virtual sample. Default is 1.
    :param particle_positions: int or list, optional
        Explicit particle positions to use. If None, positions are generated.
    :param reference_structure: str, optional
        Structure ID or path for the reference sample.
    :param reference_probe: str, optional
        Probe configuration for the reference sample.
    :param reference_parameters: dict, optional
        Additional parameters for the reference sample.
    :param clear_experiment: bool, optional
        Whether to clear the experiment before running. Default is True.
    :param run_analysis: bool, optional
        Whether to run analysis after simulation. Default is True.
    For the following parameters, pass a list of the values to sweep over,
    alternatively, pass a tuple of (min, max, nsteps) to generate linearly spaced values.
    For a single value, pass a list with one element.
    If a parameter is None, it will not be swept and will use default values
        - probe_target_type
        - probe_target_value
        - probe_target_option
        - probe_model
        - probe_fluorophore
        - probe_paratope
        - probe_conjugation_target_info
        - probe_seconday_epitope
        - peptide_motif
        - probe_distance_to_epitope
        - probe_steric_hindrance
        - probe_DoL
        - probe_wobble_theta
        - labelling_efficiency
        - structural_integrity
        - structural_integrity_small_cluster
        - structural_integrity_large_cluster
        - sample_dimensions
        - particle_orientations
        - rotation_angles
        - minimal_distance
        - pixelsize_nm
        - lateral_resolution_nm
        - axial_resolution_nm
        - psf_voxel_nm
        - depth_of_field_nm
        - exp_time

    Returns
    -------
    sweep_generator or None
        If return_generator is True, returns the sweep_generator instance.
        Otherwise, returns None.

    Notes
    -----
    - This function is the main entry point for running parameter sweeps in vLab4Mic.
    - All major simulation and analysis steps are automated.
    - For advanced usage, see the sweep_generator class and related documentation.
    """
    sweep_gen = sweep_generator(random_seed = random_seed)
    if clear_experiment:
        sweep_gen.experiment.clear_experiment()
    sweep_gen.select_structures(structures=structures)
    sweep_gen.select_probe_templates(probe_templates=probe_templates)
    sweep_gen.select_modalities(modalities=modalities)
    sweep_gen.set_output_directory(output_directory=output_directory)
    sweep_gen.set_number_of_repetitions(sweep_repetitions)
    # number of particles across sweep
    if particle_positions is not None:
        # set those positions
        sweep_gen.experiment.set_virtualsample_params(
            particle_positions=particle_positions
            )
    else:
        sweep_gen.experiment.set_virtualsample_params(
            number_of_particles=number_of_particles
            )

    sweep_gen.set_sweep_parameters(
        probe_target_type=probe_target_type,
        probe_target_value=probe_target_value,
        probe_target_option=probe_target_option,
        probe_model=probe_model,
        probe_fluorophore=probe_fluorophore,
        probe_paratope=probe_paratope,
        probe_conjugation_target_info=probe_conjugation_target_info,
        probe_seconday_epitope=probe_seconday_epitope,
        peptide_motif=peptide_motif,
        probe_distance_to_epitope=probe_distance_to_epitope,
        probe_steric_hindrance=probe_steric_hindrance,
        probe_DoL=probe_DoL,
        probe_wobble_theta=probe_wobble_theta,
        labelling_efficiency=labelling_efficiency,
        structural_integrity=structural_integrity,
        structural_integrity_small_cluster=structural_integrity_small_cluster,
        structural_integrity_large_cluster=structural_integrity_large_cluster,
        sample_dimensions=sample_dimensions,
        particle_positions=particle_positions,
        particle_orientations=particle_orientations,
        rotation_angles=rotation_angles,
        minimal_distance=minimal_distance,
        pixelsize_nm=pixelsize_nm,
        lateral_resolution_nm=lateral_resolution_nm,
        axial_resolution_nm=axial_resolution_nm,
        psf_voxel_nm=psf_voxel_nm,
        depth_of_field_nm=depth_of_field_nm,
        exp_time=exp_time,
    )
    if reference_parameters is None:
        reference_parameters = dict()
    sweep_gen.set_reference_parameters(
        reference_structure=reference_structure,
        reference_probe=reference_probe,
        **reference_parameters)
    sweep_gen.set_na_as_zero_in_plots(na_as_zero=na_as_zero)
    sweep_gen.use_default_metrics(metrics=default_metrics)         
    if custom_metrics is not None:
        sweep_gen.add_custom_analysis_metrics(
            custom_metrics=custom_metrics
            )
    if plot_parameters is not None:
        for plot_type, parameters in plot_parameters.items():
            sweep_gen.set_plot_parameters(
                plot_type=plot_type,
                **parameters)

    if run_analysis:
        sweep_gen.run_analysis(
            save=save_analysis_results, 
            plots=analysis_plots,
            output_name=output_name
            )
        if save_sweep_images:
            sweep_gen.save_images(
                output_name=output_name, 
                output_directory=output_directory
            )
    if return_generator:
        return sweep_gen


import yaml
import os
import copy
import numpy as np


def load_yaml(yaml_file: str):
    """
    Returns the associated dictionary from a yaml
    """
    with open(yaml_file, "r") as f:
        config_dictionary = yaml.safe_load(f)
    return config_dictionary


def save_yaml(data=None, name="NONAME", output_directory=None):
    """
    Saves a dictionary to a yaml file
    """
    file_name = name + ".yml"
    file_path = os.path.join(output_directory, file_name)
    #yaml.Dumper.ignore_aliases = lambda self, data: True
    with open(file_path, "w") as f:
        yaml.safe_dump(data, f, default_flow_style=False)


from vlab4mic import workflows, experiments
from vlab4mic.utils import data_format
import pytest
import copy


def test_simple_imaging_system():
    imager, _ = experiments.build_virtual_microscope()
    assert imager.get_absolute_reference_point().shape == (1, 3)


def test_get_raw_volume(experiment_7r5k_base):
    experiment_7r5k_base
    images_volumes, beads, img_noiseless, b_noiseless = experiment_7r5k_base.imager.generate_imaging(
        modality="SMLM", convolution_type="raw_volume", exp_time=0.001
    )
    assert images_volumes["raw_volume"][0].shape == (200, 200, 150)


def test_multi_imaging_system():
    imager, _ = experiments.build_virtual_microscope(
        multimodal=["STED", "Confocal"]
    )
    assert imager.get_absolute_reference_point().shape == (1, 3)


def test_image_from_field(configuration_directory, gt_structural_model_field):
    configuration_path = configuration_directory
    selected_mods = [
        "STED",
    ]
    imgs, imgs_noiseless, experiment_test = experiments.image_vsample(
        vsample=gt_structural_model_field,
        run_simulation=True,
        save=True,
        multimodal=selected_mods,
    )
    assert experiment_test.imager.get_absolute_reference_point().shape == (
        1,
        3,
    )


def test_imager_optional_methods(experiment_7r5k_base):
    imager = copy.deepcopy(experiment_7r5k_base.imager)
    imager.set_experiment_name("test")
    imager.set_focus(0.23)
    imager.set_roi_position(x=0.1, y=0.2)
    imager.set_roi_sizes(x=2, y=2)
    imager.recenter_roi()
    #imager.show_field()



import numpy as np


def add_poisson_noise(imagestack):
    # print("adding poisson noise")
    noisy = np.random.poisson(imagestack)
    return noisy


def add_gaussian_noise(imagestack, sigma):
    # print(f"adding gaussian noise: sigma = {sigma}")
    gnoise = np.random.normal(0, sigma, size=imagestack.shape)
    noisy = imagestack + gnoise
    return noisy


def add_binomial_noise(imagestack, p=1.0):
    # print(f"adding binomial noise: p = {p}")
    # int_imagestack = np.floor(imagestack).astype("int64")
    imagestack = np.floor(imagestack).astype(np.int32)
    imstack_binom = np.random.binomial(imagestack, p)
    return imstack_binom


def add_gamma_noise(imagestack, g=1.0):
    # print(f"adding gamma noise: g = {g}")
    amplified = np.random.gamma(imagestack, scale=g)
    return amplified


def add_conversion_factor(imagestack, adu: int = 1):
    """
    Emulates the digitalisation process for generating
    pixel values as integers
    """
    # print(f"Using conversion factor: ADU = {adu}")
    adu_stack = np.floor(imagestack / adu).astype(np.int32)
    return adu_stack


def add_integer_baselevel(imagestack, bl: int = 0):
    """
    Adds a constant value as integer.
    As the input data should be integers as well it is
    explicitly casted as integer
    """
    stack_with_bl = (imagestack.astype(np.int32)) + bl
    return stack_with_bl


def add_image_noise(noise_type: str, stack, **kwargs):
    """
    Add noise on a image-based scheme. This models a EMCCD
    or any noise that is not pixel-dependent
    """
    if noise_type == "binomial":
        noisy = add_binomial_noise(stack, **kwargs)
    elif noise_type == "gaussian":
        noisy = add_gaussian_noise(stack, **kwargs)
    elif noise_type == "gamma":
        noisy = add_gamma_noise(stack, **kwargs)
    elif noise_type == "poisson":
        noisy = add_poisson_noise(stack)
    elif noise_type == "conversion":
        noisy = add_conversion_factor(stack, **kwargs)
    elif noise_type == "baselevel":
        noisy = add_integer_baselevel(stack, **kwargs)
    else:
        noisy = None
    return noisy



import pytest
from vlab4mic import workflows, experiments
from vlab4mic.utils import data_format


def test_primary_secondary_fromExperiment():
    sample, test_experiment = experiments.generate_virtual_sample(
        clear_probes=True
    )
    test_experiment.remove_probes()
    test_experiment.add_probe(
        probe_template="Linker",
        probe_target_type="Sequence",
        probe_target_value="EQATETQ",
    )
    test_experiment.add_probe(
        probe_template="Antibody",
        probe_target_type="Primary",
        probe_target_value="Linker",
    )
    test_experiment.build(modules=["particle"])
    assert test_experiment.particle is not None
    assert test_experiment.exported_coordinate_field["field_emitters"] != {}


def test_primaryAB_secondaryAB_fromExperiment():
    modalities = ["STED", "SMLM",]
    primary = dict(
        probe_template = "Antibody",
        probe_name="Primary-clathrin",
        probe_target_type = "Sequence",
        probe_target_value = "EQATETQ",
    )
    secondary = dict(
        probe_template = "Antibody",
        probe_name="Secondary-clathrin",
        probe_target_type = "Primary",
        probe_target_value = "Primary-clathrin"
    )
    image_outputs4, image_outputs_noiseless4, experiment = experiments.image_vsample(
        structure = "1XI5",
        primary_probe = primary,
        secondary_probe = secondary,
        multimodal=modalities,
        clear_experiment=True,
        run_simulation=True
    )

    assert len(list(experiment.probe_parameters.keys())) == 2
    assert len(list(experiment.particle.secondary.keys())) == 1 
    assert experiment.particle.sequential_labelling == True




import tifffile as tif
from datetime import datetime
import numpy as np
import os


def write_tif(image_stack: np.ndarray, dirname: str, notes: str):
    now = datetime.now()  # dd/mm/YY H:M:S
    dt_string = now.strftime("%Y%m%d") + "_"
    # adjust intensities to match pixel depth
    if image_stack.max() < (2**16 - 1):
        image_stack = image_stack.astype(np.uint16)
    else:
        image_stack = image_stack.astype(np.uint32)
    filename = dirname + dt_string + notes + ".tif"
    current_files = os.listdir(dirname)
    if filename.split("/")[-1] in current_files:
        count = 1
        new_filename = (
            dirname
            + dt_string
            + notes
            + "_"
            + str(count)
            + ".tif"
        )
        while new_filename.split("/")[-1] in current_files:
            count += 1
            new_filename = (
                dirname
                + dt_string
                + notes
                + "_"
                + str(count)
                + ".tif"
            )
        filename = new_filename
    print(f"image saved in {filename}")
    tif.imwrite(filename, image_stack, bigtiff=True)



import os
from pathlib import Path

import requests
import yaml
from tqdm import tqdm

headers = {
    "User-Agent": (
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) "
        "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/102.0.0.0 Safari/537.36"
    ),
    "Accept": "application/json, text/plain, */*",
    "Accept-Language": "en-GB",
    # 'Accept-Encoding': 'gzip, deflate, br',
    "Referer": "https://henriqueslab.org/",
    "Origin": "https://henriqueslab.org",
    "Connection": "keep-alive",
    "Sec-Fetch-Dest": "empty",
    "Sec-Fetch-Mode": "cors",
    "Sec-Fetch-Site": "same-site",
    "Pragma": "no-cache",
    "Cache-Control": "no-cache",
}


def download_suggested_structures(data_path: str = "data") -> None:
    """
    Downloads PDB *.cif files for the suggested structures in the given path.

    Args:
        data_path: The path to the data directory. Defaults to "data".

    Returns:
        None
    """

    structures_path = Path(data_path) / "structures"

    for filename in structures_path.glob("*.yaml"):
        if filename.stem == "_template":
            continue

        cifs_filename = filename.with_suffix(".cif")

        with open(filename, "r") as f:
            data = yaml.load(f, Loader=yaml.FullLoader)

        id = data["id"]
        title = data["title"]

        if cifs_filename.exists():
            print(f"{title} - [{cifs_filename}] already exists. Skipping...")
            continue

        print(f"Downloading {title} - [{cifs_filename}]...")
        url = f"https://files.rcsb.org/download/{id}.cif"
        download_file(url, filename.with_suffix(".cif"))


def download_file(url: str, fname: str, chunk_size: int = 1024) -> None:
    """Downloads a file from the given URL and saves it to the specified file name.

    Args:
        url: The URL of the file to download.
        fname: The file name to save the downloaded file as.
        chunk_size: The size of the chunks to download the file in. Defaults to 1024.

    Returns:
        None
    """
    resp = requests.get(url, stream=True, headers=headers)
    total = int(resp.headers.get("content-length", 0))

    desc = os.path.basename(fname)
    # Truncate description if too long.
    if len(desc) > 50:
        desc = desc[:30] + "..."

    with (
        open(fname, "wb") as file,
        tqdm(
            desc=desc,
            total=total,
            unit="iB",
            unit_scale=True,
            unit_divisor=1024,
        ) as bar,
    ):
        for data in resp.iter_content(chunk_size=chunk_size):
            size = file.write(data)
            bar.update(size)


def verify_structure(structureid: str, structure_dir: str) -> None:
    """
    Checks if .cif exist alreadly. Otherwise, download it.
    Downloads PDB *.cif files for the suggested structures in the given path.

    Args:
        data_path: The path to the data directory. Defaults to "data".

    Returns:
        None
    """

    # get CIF path
    cif_name = structureid + ".cif"
    cif_file = os.path.join(structure_dir, cif_name)
    cif_file_path = Path(cif_file)
    if cif_file_path.exists():
        print(f"{structureid} already exists. Skipping...")
        return cif_file_path
    else:
        print(f"Downloading {structureid}...")
        url = f"https://files.rcsb.org/download/{structureid}.cif"
        download_file(url, cif_file_path)
        return cif_file_path



import matplotlib.pyplot as plt
import numpy as np
import yaml
import copy

from ..utils.visualisation.matplotlib_plots import (
    add_ax_scatter,
    draw1nomral_segment,
)
from ..utils.data_format.visualisation import format_coordinates
from ..utils.io.yaml_functions import load_yaml


class Label:
    def __init__(self):
        self.params = {}
        self.params["label_name"] = None
        self.params["labeling_efficiency"] = 1
        self.params["label_type"] = None
        self.params["fluorophore"] = None
        self.params["excitation"] = None
        self.params["emission"] = None
        self.params["atoms"] = None
        self.params["residues"] = None
        self.params["chains"] = None
        self.params["position"] = None
        self.params["plotcolour"] = None

        self.params["emitters_coords"] = None
        self.params["axis"] = dict(pivot=[0, 0, 0], direction=[0, 0, 1])

        self.params["target_sequence"] = None
        self.params["summary_method"] = "average"
        self.params["length"] = 0
        self.conjugation = dict()
        self.model = dict()
        self.binding = dict()
        self.epitope = dict()

    def set_params(self, **kwargs):
        """
        Method used to set the parameters using keyword arguments.
        :param kwargs: same as self.params.keys()
        """
        for key, value in kwargs.items():
            self.params[key] = value

    def _set_model_params(self, ID=None, format=None, database=None, **kwargs):
        self.model["ID"] = ID
        self.model["format"] = format
        self.model["database"] = database
        for key, value in kwargs.items():
            self.model[key] = value

    def _set_conjugation_params(self, target: dict, DoL=None, **kwargs):
        self.conjugation["target"] = target
        self.conjugation["DoL"] = DoL
        for key, value in kwargs.items():
            self.conjugation[key] = value

    def _set_binding_params(
        self,
        efficiency,
        orientation,
        wobble_range: dict,
        distance: dict,
        paratope: str,
        **kwargs,
    ):

        self.binding["efficiency"] = efficiency
        self.binding["orientation"] = orientation
        self.binding["wobble_range"] = wobble_range
        self.binding["distance"] = distance
        self.binding["paratope"] = paratope
        for key, value in kwargs.items():
            self.binding[key] = value

    def _set_epitope_params(
        self, target=None, normals=None, site=None, **kwargs
    ):
        self.epitope["target"] = target
        self.epitope["normals"] = normals
        self.epitope["site"] = site

    def set_axis(self, pivot: list, direction: list):
        self.params["axis"]["pivot"] = pivot
        self.params["axis"]["direction"] = direction

    def set_fluorophore(self, fluo_id):
        self.params["fluorophore"] = fluo_id

    def get_name(self):
        return self.params["label_name"]

    def get_label_type(self):
        return self.params["label_type"]

    def get_target_type(self):
        return self.params["target"]["type"]

    def get_plotcolour(self):
        return self.params["plotcolour"]

    def get_fluorophore(self):
        return self.params["fluorophore"]

    def get_efficiency(self):
        return self.params["labeling_efficiency"]

    def get_param(self, param: str):
        return self.params[param]

    def generate_linker(self):
        if self.params["length"] > 0:
            single_emitter = np.array([0, 0, self.params["length"]])
            single_emitter = single_emitter[np.newaxis, :]
            self.set_emitters(single_emitter)
        else:
            print("No linker created due to linker lenght == 0")

    def set_target_sequence(self, sequence: str):
        self.params["target_sequence"] = sequence

    def set_emitters(self, emitters):
        if (np.shape(emitters))[0] < 1:
            print("input need at least 1 points")
        else:
            self.params["emitters_coords"] = emitters

    def gen_labeling_entity(self):
        labeling_emitters = copy.copy(self.params["emitters_coords"])
        p1 = np.array(self.params["axis"]["pivot"])

        # Safety check for axis direction - use default if None
        direction = self.params["axis"]["direction"]
        if direction is None:
            direction = [0, 0, 1]  # Default axis direction

        p2 = p1 + np.array(direction)
        pivots = np.array([p1, p2])
        # print(f"pivots are: {pivots}")
        if len(labeling_emitters) < 1:
            print("there are no emitters specified")
        else:
            labeling_entity = np.vstack([pivots, labeling_emitters])
        return labeling_entity

    def export_as_file(
        self, filename="customlabel.yaml", write_dir=""
    ):
        writing_dir = write_dir + filename
        print(f"Writing file in: {writing_dir}")
        # define parameters to write
        if self.params["emitters_coords"] is None:
            coordinates = None
        else:
            coordinates = self.gen_labeling_entity()
            coordinates = coordinates.tolist()
        label_params = dict(
            label_type=self.get_param("label_type"),
            label_name="NAME",
            labeling_efficiency=self.get_efficiency(),
            coordinates=coordinates,
            target_sequence=self.get_param("target_sequence"),
            plotcolour=None,
            fluorophore=self.get_param("fluorophore"),
        )
        yaml_dict = dict(
            labels=[
                label_params,
            ]
        )
        with open(writing_dir, "w") as file:
            yaml.dump(yaml_dict, file)

    def plot_emitters(
        self,
        view_init=[0, 0, 0],
        axesoff=True,
        show_axis=True,
        return_plot=False,
    ):
        if self.params["emitters_coords"] is None:
            print(
                f"Label type is {self.params['label_type']} and does not contain "
                "emitter coordinates. To include them manually use method set_emitters"
            )
        else:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection="3d")
            add_ax_scatter(
                ax, format_coordinates(self.params["emitters_coords"])
            )
            ax.view_init(
                elev=view_init[0], azim=view_init[1], roll=view_init[2]
            )
            if show_axis:
                draw1nomral_segment(
                    self.params["axis"], ax, lenght=150, colors=["g", "y"]
                )
            ax = plt.gca()
            emitters = self.params["emitters_coords"]
            minimum = np.min(emitters)
            maximum = np.max(emitters)
            ax.set_xlim(minimum, maximum)
            ax.set_ylim(minimum, maximum)
            ax.set_zlim(minimum, maximum)
            if axesoff:
                ax.set_axis_off()
            ax.set_box_aspect(
                [
                    ub - lb
                    for lb, ub in (getattr(ax, f"get_{a}lim")() for a in "xyz")
                ]
            )
            if return_plot:
                plt.close()
                return fig
            else:
                fig.show

    def get_notNone_params(self):
        for key, value in self.params.items():
            if value is not None:
                print(f"{key}: {value}")


def construct_label(
    label_config_dictionary: None,
    lab_eff: float = None,
    target_info=None,
):
    """
    Construct an object of class Label.
    Assumes there exist a configuration file

    Args:
        label_id: (string) ID of Label configuration file
        fluorophore_id: (string) ID of fluorophore configuration file
    Output:
        label: (Label) Object of class Label()
        label_params: (dictionary)
        fluorophore_params: (dictionary)
    """
    label_params = copy.deepcopy(label_config_dictionary)

    # keynames reserved for parameters that can be iteratet over
    if "model_ID" in label_config_dictionary.keys():
        label_params["model"]["ID"] = label_config_dictionary["model_ID"]
    if "distance_to_epitope" in label_config_dictionary.keys():
        label_params["binding"]["distance"]["to_target"] = label_params[
            "distance_to_epitope"
        ]
    else:
        print("No distance to epitope provided, using default value.")
    if "distance_between_epitope" in label_config_dictionary.keys():
        label_params["binding"]["distance"]["between_targets"] = label_config_dictionary[
            "distance_between_epitope"
        ]
    if "paratope" in label_config_dictionary.keys():
        label_params["binding"]["paratope"] = label_config_dictionary["paratope"]
    if "conjugation_target_info" in label_config_dictionary.keys():
        label_params["conjugation_sites"]["target"] = label_config_dictionary[
            "conjugation_target_info"
        ]
    #if "probe_DoL" in label_params.keys():
    #    label_params["conjugation_sites"]["DoL"] = kwargs[
    #        "probe_DoL"
    #    ]
    #else:
    #    print("No DoL provided, using default value. ########################")
    if "epitope_target_info" in label_config_dictionary.keys():
        label_params["epitope"]["target"] = label_config_dictionary["epitope_target_info"]
    if "wobble_theta" in label_params.keys():
        label_params["binding"]["wobble_range"]["theta"] = label_params[
            "wobble_theta"
        ]
    else:
        label_params["binding"]["wobble_range"]["theta"] = None
    ######## information about target
    if target_info:
        # expect label_params to have empty values on target type and value
        label_params["target"]["type"] = target_info["type"]
        label_params["target"]["value"] = target_info["value"]
    if label_params["target"]["type"] == "Atom_residue":
        label_params["atoms"] = [
            label_params["target"]["value"]["atoms"],
        ]
        label_params["residues"] = [
            label_params["target"]["value"]["residues"],
        ]
        try:
            label_params["position"] = label_params["target"]["value"][
                "position"
            ]
        except:
            label_params["position"] = None
    ######## Building the labelling entity: anribody, linker, direct...
    if "fluorophore_id" not in label_params.keys():
        fluorophore_id = "AF647"
    else:
        fluorophore_id = label_params["fluorophore_id"]
    label_params["fluorophore"] = fluorophore_id
    if lab_eff is not None:
        label_params["labelling_efficiency"] = lab_eff
    label = Label()
    label.set_params(**label_params)
    label.set_fluorophore(fluorophore_id)
    # check whether there is enough structural information for the labelling entity
    if (
        label_params["model"]["ID"]
        and label_params["conjugation_sites"]["target"]["type"]
    ):
        # building probe from PDB/CIF
        print("antibody")
        label._set_model_params(**label_params["model"])
        label._set_conjugation_params(**label_params["conjugation_sites"])
        label._set_binding_params(**label_params["binding"])
        if "epitope" in label_params.keys():
            label._set_epitope_params(**label_params["epitope"])
        else:
            label._set_epitope_params()
        # label.emitters_from_PDBCIF(**label_params)
        # label_params["coordinates"] = label.gen_labeling_entity()
    elif label_params["binding"]["distance"]["to_target"]:
        # not from a PDB, checking if at least distance to make a rigid linker
        print("rigid linker")
        label.set_params(
            length=label_params["binding"]["distance"]["to_target"]
        )
        label.set_axis(
            pivot=[0, 0, 0], direction=label_params["binding"]["orientation"]
        )
        label.generate_linker()
        label_params["coordinates"] = label.gen_labeling_entity()
    else:
        print("direct")
    # if none of these conditions exist, it will be treated as direct
    return label, label_params
